//  Copyright (C) 2000  MandrakeSoft S.A.
//
//    MandrakeSoft S.A.
//    43, rue d'Aboukir
//    75002 Paris - France
//    http://www.linux-mandrake.com/
//    http://www.mandrakesoft.com/
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA



#include "bochs.h"
#include "gui/bitmaps/floppya.h"
#include "gui/bitmaps/floppyb.h"
#include "gui/bitmaps/mouse.h"
#include "gui/bitmaps/reset.h"
#include "gui/bitmaps/power.h"
#include "gui/bitmaps/snapshot.h"
#ifdef macintosh
#  include <Disks.h>
#endif



bx_gui_c   bx_gui;

#define BX_GUI_THIS bx_gui.


  void
bx_gui_c::init(int argc, char **argv, unsigned tilewidth, unsigned tileheight)
{
  specific_init(&bx_gui, argc, argv, tilewidth, tileheight, BX_HEADER_BAR_Y);

  // Define some bitmaps to use in the headerbar
  BX_GUI_THIS floppyA_bmap_id = create_bitmap(bx_floppya_bmap,
                          BX_FLOPPYA_BMAP_X, BX_FLOPPYA_BMAP_Y);
  BX_GUI_THIS floppyA_eject_bmap_id = create_bitmap(bx_floppya_eject_bmap,
                          BX_FLOPPYA_BMAP_X, BX_FLOPPYA_BMAP_Y);
  BX_GUI_THIS floppyB_bmap_id = create_bitmap(bx_floppyb_bmap,
                          BX_FLOPPYB_BMAP_X, BX_FLOPPYB_BMAP_Y);
  BX_GUI_THIS floppyB_eject_bmap_id = create_bitmap(bx_floppyb_eject_bmap,
                          BX_FLOPPYB_BMAP_X, BX_FLOPPYB_BMAP_Y);
  BX_GUI_THIS mouse_bmap_id = create_bitmap(bx_mouse_bmap,
                          BX_MOUSE_BMAP_X, BX_MOUSE_BMAP_Y);
  BX_GUI_THIS nomouse_bmap_id = create_bitmap(bx_nomouse_bmap,
                          BX_MOUSE_BMAP_X, BX_MOUSE_BMAP_Y);


  BX_GUI_THIS power_bmap_id = create_bitmap(bx_power_bmap, BX_POWER_BMAP_X, BX_POWER_BMAP_Y);
  BX_GUI_THIS reset_bmap_id = create_bitmap(bx_reset_bmap, BX_RESET_BMAP_X, BX_RESET_BMAP_Y);
  BX_GUI_THIS snapshot_bmap_id = create_bitmap(bx_snapshot_bmap, BX_SNAPSHOT_BMAP_X, BX_SNAPSHOT_BMAP_Y);


  // Add the initial bitmaps to the headerbar, and enable callback routine, for use
  // when that bitmap is clicked on

  // Floppy A:
  BX_GUI_THIS floppyA_status = bx_devices.floppy->get_media_status(0);
#ifndef __OS2__
  if (BX_GUI_THIS floppyA_status)
    BX_GUI_THIS floppyA_hbar_id = headerbar_bitmap(BX_GUI_THIS floppyA_bmap_id,
                          BX_GRAVITY_LEFT, floppyA_handler);
  else
    BX_GUI_THIS floppyA_hbar_id = headerbar_bitmap(BX_GUI_THIS floppyA_eject_bmap_id,
                          BX_GRAVITY_LEFT, floppyA_handler);
#endif

  // Floppy B:
  BX_GUI_THIS floppyB_status = bx_devices.floppy->get_media_status(1);
#ifndef __OS2__
  if (BX_GUI_THIS floppyB_status)
    BX_GUI_THIS floppyB_hbar_id = headerbar_bitmap(BX_GUI_THIS floppyB_bmap_id,
                          BX_GRAVITY_LEFT, floppyB_handler);
  else
    BX_GUI_THIS floppyB_hbar_id = headerbar_bitmap(BX_GUI_THIS floppyB_eject_bmap_id,
                          BX_GRAVITY_LEFT, floppyB_handler);
#endif

  // Mouse button
  BX_GUI_THIS mouse_status = gui_get_mouse_enable();
  if (BX_GUI_THIS mouse_status)
    BX_GUI_THIS mouse_hbar_id = headerbar_bitmap(BX_GUI_THIS mouse_bmap_id,
                          BX_GRAVITY_LEFT, mouse_handler);
  else
    BX_GUI_THIS mouse_hbar_id = headerbar_bitmap(BX_GUI_THIS nomouse_bmap_id,
                          BX_GRAVITY_LEFT, mouse_handler);

  // Power button
  BX_GUI_THIS power_hbar_id = headerbar_bitmap(BX_GUI_THIS power_bmap_id,
                          BX_GRAVITY_RIGHT, power_handler);
  // Reset button
  BX_GUI_THIS reset_hbar_id = headerbar_bitmap(BX_GUI_THIS reset_bmap_id,
                          BX_GRAVITY_RIGHT, reset_handler);
  // Snapshot button
  BX_GUI_THIS snapshot_hbar_id = headerbar_bitmap(BX_GUI_THIS snapshot_bmap_id,
                          BX_GRAVITY_RIGHT, snapshot_handler);

  show_headerbar();

  bx_g2h.aquire_channel(bx_gui.direct_access,0x1010101);
}


  Boolean
bx_gui_c::floppyA_handler(Boolean bQueryStatus)
{
  unsigned new_status;

  if ( bQueryStatus )
    {
    BX_GUI_THIS floppyA_status = bx_devices.floppy->get_media_status(0);
    return BX_GUI_THIS floppyA_status ? false : true ;
    }

  new_status = bx_devices.floppy->set_media_status(0, !BX_GUI_THIS floppyA_status);
  if (new_status == BX_GUI_THIS floppyA_status)
    return BX_GUI_THIS floppyA_status ? false : true ;

  BX_GUI_THIS floppyA_status = new_status;
  if (BX_GUI_THIS floppyA_status)
    replace_bitmap(BX_GUI_THIS floppyA_hbar_id, BX_GUI_THIS floppyA_bmap_id);
  else {
#ifdef macintosh
    // If we are using the Mac floppy driver, eject the disk
    // from the floppy drive
    if (!strcmp(bx_options.floppya.path, SuperDrive))
      DiskEject(1);
#endif
    replace_bitmap(BX_GUI_THIS floppyA_hbar_id, BX_GUI_THIS floppyA_eject_bmap_id);
    }

  return BX_GUI_THIS floppyA_status ? false : true ;
}

  Boolean
bx_gui_c::floppyB_handler(Boolean bQueryStatus)
{
  unsigned new_status;

  if ( bQueryStatus )
    {
    BX_GUI_THIS floppyB_status = bx_devices.floppy->get_media_status(1);
    return BX_GUI_THIS floppyB_status ? false : true;
    }

  new_status = bx_devices.floppy->set_media_status(1, !BX_GUI_THIS floppyB_status);
  if (new_status == BX_GUI_THIS floppyB_status)
    return BX_GUI_THIS floppyB_status ? false : true;

  BX_GUI_THIS floppyB_status = new_status;
  if (BX_GUI_THIS floppyB_status)
    replace_bitmap(BX_GUI_THIS floppyB_hbar_id, BX_GUI_THIS floppyB_bmap_id);
  else
    replace_bitmap(BX_GUI_THIS floppyB_hbar_id, BX_GUI_THIS floppyB_eject_bmap_id);

  return BX_GUI_THIS floppyB_status ? false : true;
}

  Boolean
bx_gui_c::printer_handler(Boolean bQueryStatus, unsigned port)
{
  if ( bQueryStatus )
    return BX_GUI_THIS printer_status[port] ? true : false ;

  BX_GUI_THIS printer_status[port] = !BX_GUI_THIS printer_status[port];
  bx_devices.parallel->set_status(port, BX_GUI_THIS printer_status[port]);
  return BX_GUI_THIS printer_status[port] ? true : false ;
}

  Boolean
bx_gui_c::cdrom_handler(Boolean bQueryStatus)
{
  if ( bQueryStatus )
    {
    BX_GUI_THIS cdrom_status = bx_devices.hard_drive->get_cdrom_status();
    return BX_GUI_THIS cdrom_status ? false : true;
    }

  BX_GUI_THIS cdrom_status = !BX_GUI_THIS cdrom_status;
  bx_devices.hard_drive->set_cdrom_status(BX_GUI_THIS cdrom_status);
  return BX_GUI_THIS cdrom_status ? false : true;
}

  void
bx_gui_c::reset_handler(void)
{
  fprintf(stderr, "# RESET callback (unimplemented).\n");
}

  void
bx_gui_c::power_handler(void)
{
  bx_panic("POWER button turned off.\n");
}

  void
bx_gui_c::snapshot_handler(void)
{
  fprintf(stderr, "# SNAPSHOT callback (unimplemented).\n");
}

  void
bx_gui_c::mouse_handler(void)
{
  BX_GUI_THIS mouse_status = ! BX_GUI_THIS mouse_status;

  if (BX_GUI_THIS mouse_status)
    replace_bitmap(BX_GUI_THIS mouse_hbar_id, BX_GUI_THIS mouse_bmap_id);
  else
    replace_bitmap(BX_GUI_THIS mouse_hbar_id, BX_GUI_THIS nomouse_bmap_id);

  gui_set_mouse_enable(BX_GUI_THIS mouse_status);
}

  Boolean
bx_gui_c::gui_get_mouse_enable(void)
{
  return(bx_options.mouse_enabled);
}

  void
bx_gui_c::gui_set_mouse_enable(Boolean val)
{
  bx_options.mouse_enabled = val;
}

#ifdef __OS2__
#pragma pack(1)
#endif

#define BX_GUI_DA_DIMENSION     'D'
#define BX_GUI_DA_BLIT          'B'
#define BX_GUI_DA_PALETTE       'P'

typedef struct _direct_access_struct
{
   Bit8u type;
   union
   {
     struct
     {
       Bit16u cx;
       Bit16u cy;
     } dimension;
     struct
     {
       Bit16u x;
       Bit16u y;
       Bit8u tile[16][16];
     } blit;
     struct
     {
       Bit16u count;
       struct
       {
         Bit16u red;
         Bit16u green;
         Bit16u blue;
       } entry[256];
     } palette;
   } packet;
} BX_DIRECT_ACCESS_STRUCT;

  void
bx_gui_c::direct_access(Bit32u *pCallbackStruct)
{
  BX_DIRECT_ACCESS_STRUCT *pdas;
  unsigned i;

  pdas = (BX_DIRECT_ACCESS_STRUCT *) calloc(1,pCallbackStruct[3]);
  BX_MEM.read_physical( pCallbackStruct[2], pCallbackStruct[3], pdas );

  switch ( pdas->type )
  {
    case BX_GUI_DA_DIMENSION:
      bx_gui.dimension_update( pdas->packet.dimension.cx,
                               pdas->packet.dimension.cy );
      break;
    case BX_GUI_DA_BLIT:
      bx_gui.graphics_tile_update( (Bit8u *) pdas->packet.blit.tile,
                                   pdas->packet.blit.x,
                                   pdas->packet.blit.y );
      break;
    case BX_GUI_DA_PALETTE:
      for ( i = 0; i < pdas->packet.palette.count; i++ )
        bx_gui.palette_change( i,
                               pdas->packet.palette.entry[i].red,
                               pdas->packet.palette.entry[i].green,
                               pdas->packet.palette.entry[i].blue );
      break;
    default:
      bx_printf( "bx_gui.direct_access: unknown packet type %x\n", pdas->type );
  }
}
