// Copyright Notice
// ================
// BOCHS is Copyright 1994-1999 by Kevin P. Lawton.
//
// BOCHS is commercial software.
//
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
//
//    US Mail:  Kevin Lawton
//              528 Lexington St.
//              Waltham, MA 02452-3028
//
//    EMail:    bochs@world.std.com



#include "bochs.h"

#define INCL_DOS
#define INCL_DOSPROCESS
#define INCL_DOSMODULEMGR
#include <os2.h>


HMODULE hModule;

void (* APIENTRY gui_dll_init)( bx_options_t *poption,
				bx_mem_c *pmem,
				Boolean (*floppyA_handler)(Boolean),
				Boolean (*floppyB_handler)(Boolean),
				Boolean (*cdrom_handler)(Boolean),
				Boolean (*printer_handler)(Boolean, unsigned int),
				void (*put_scancode)(unsigned char *, int),
				void (*mouse_motion)(int,int,unsigned int),
				HMODULE hModule );
void (* APIENTRY gui_specific_init)(bx_gui_c *th,
				int argc,
				char **argv,
				unsigned tilewidth,
				unsigned tileheight,
				unsigned headerbar_y);
void (* APIENTRY gui_handle_events)(void);
void (* APIENTRY gui_text_update)(Bit8u *old_text,
				Bit8u *new_text,
				unsigned long cursor_x,
				unsigned long cursor_y,
				unsigned nrows);
Boolean (* APIENTRY gui_palette_change)(unsigned index,
				unsigned red,
				unsigned green,
				unsigned blue);
void (* APIENTRY gui_graphics_tile_update)(Bit8u *tile,
				unsigned x0,
				unsigned y0);
void (* APIENTRY gui_dimension_update)(unsigned x,
				unsigned y);
void (* APIENTRY gui_exit)(void);



// This file defines stubs for the GUI interface, which is a
// place to start if you want to port bochs to a platform, for
// which there is no support for your native GUI, or if you want to compile
// bochs without any native GUI support (no output window or
// keyboard input will be possible).
// Look in 'x.cc', 'beos.cc', and 'win32.cc' for specific
// implementations of this interface.  -Kevin



// ::SPECIFIC_INIT()
//
// Called from gui.cc, once upon program startup, to allow for the
// specific GUI code (X11, BeOS, ...) to be initialized.
//
// th: a 'this' pointer to the gui class.  If a function external to the
//     class needs access, store this pointer and use later.
// argc, argv: not used right now, but the intention is to pass native GUI
//     specific options from the command line.  (X11 options, BeOS options,...)
//
// tilewidth, tileheight: for optimization, graphics_tile_update() passes
//     only updated regions of the screen to the gui code to be redrawn.
//     These define the dimensions of a region (tile).
// headerbar_y:  A headerbar (toolbar) is display on the top of the
//     VGA window, showing floppy status, and other information.  It
//     always assumes the width of the current VGA mode width, but
//     it's height is defined by this parameter.

void bx_gui_c::specific_init(bx_gui_c *th, int argc, char **argv, unsigned tilewidth, unsigned tileheight, unsigned headerbar_y)
{
  CHAR szLoadError[256];
  ULONG rc;

  switch ( bx_options.videomode )
     {
     case BX_VIDEO_PM_WINDOW:
        rc = DosLoadModule(szLoadError, 256, "os2pm.dll", &hModule );
        break;
     case BX_VIDEO_DIVE_WINDOW:
        rc = DosLoadModule(szLoadError, 256, "os2dv.dll", &hModule );
        break;
     case BX_VIDEO_DIRECT:
     case BX_VIDEO_FULLSCREEN:
        rc = DosLoadModule(szLoadError, 256, "os2fs.dll", &hModule );
        break;
     }

  if ( rc )
     {
     bx_printf( "can't load gui dll\n" );
     return;
     }

  DosQueryProcAddr( hModule, 0, "gui_dll_init",             (PFN*) &gui_dll_init );
  DosQueryProcAddr( hModule, 0, "gui_specific_init",        (PFN*) &gui_specific_init );
  DosQueryProcAddr( hModule, 0, "gui_handle_events",        (PFN*) &gui_handle_events );
  DosQueryProcAddr( hModule, 0, "gui_text_update",          (PFN*) &gui_text_update );
  DosQueryProcAddr( hModule, 0, "gui_palette_change",       (PFN*) &gui_palette_change );
  DosQueryProcAddr( hModule, 0, "gui_graphics_tile_update", (PFN*) &gui_graphics_tile_update );
  DosQueryProcAddr( hModule, 0, "gui_dimension_update",     (PFN*) &gui_dimension_update );
  DosQueryProcAddr( hModule, 0, "gui_exit",                 (PFN*) &gui_exit );

  if ( !gui_dll_init || !gui_specific_init )
     {
     bx_printf( "gui dll entry not found for gui_dll_init or gui_specific_init\n" );
     return;
     }

  gui_dll_init(
	&bx_options,
	&bx_mem,
	floppyA_handler,
	floppyB_handler,
        cdrom_handler,
	printer_handler,
	bx_devices.keyboard->put_scancode,
	bx_devices.keyboard->mouse_motion,
        hModule );

  gui_specific_init(th,argc,argv,tilewidth,tileheight,headerbar_y);
}


// ::HANDLE_EVENTS()
//
// Called periodically (vga_update_interval in .bochsrc) so the
// the gui code can poll for keyboard, mouse, and other
// relevant events.

void bx_gui_c::handle_events(void)
{
  if ( !gui_handle_events )
     {
     bx_printf( "gui dll entry not found for gui_handle_events\n" );
     return;
     }

   gui_handle_events();
}


// ::FLUSH()
//
// Called periodically, requesting that the gui code flush all pending
// screen update requests.

void bx_gui_c::flush(void)
{
}


// ::CLEAR_SCREEN()
//
// Called to request that the VGA region is cleared.  Don't
// clear the area that defines the headerbar.

void bx_gui_c::clear_screen(void)
{
}



// ::TEXT_UPDATE()
//
// Called in a VGA text mode, to update the screen with
// new content.
//
// old_text: array of character/attributes making up the contents
//           of the screen from the last call.  See below
// new_text: array of character/attributes making up the current
//           contents, which should now be displayed.  See below
//
// format of old_text & new_text: each is 4000 bytes long.
//     This represents 80 characters wide by 25 high, with
//     each character being 2 bytes.  The first by is the
//     character value, the second is the attribute byte.
//     I currently don't handle the attribute byte.
//
// cursor_x: new x location of cursor
// cursor_y: new y location of cursor

void bx_gui_c::text_update(Bit8u *old_text, Bit8u *new_text, unsigned long cursor_x, unsigned long cursor_y, unsigned nrows)
{
  if ( !gui_text_update )
     {
     bx_printf( "gui dll entry not found for gui_text_update\n" );
     return;
     }
  gui_text_update(old_text,new_text,cursor_x,cursor_y,nrows);
}


// ::PALETTE_CHANGE()
//
// Allocate a color in the native GUI, for this color, and put
// it in the colormap location 'index'.
// returns: 0=no screen update needed (color map change has direct effect)
//          1=screen updated needed (redraw using current colormap)

Boolean bx_gui_c::palette_change(unsigned index, unsigned red, unsigned green, unsigned blue)
{
  if ( !gui_palette_change )
     {
     bx_printf( "gui dll entry not found for gui_palette_change\n" );
     return false;
     }
  return gui_palette_change(index,red,green,blue);
}


// ::GRAPHICS_TILE_UPDATE()
//
// Called to request that a tile of graphics be drawn to the
// screen, since info in this region has changed.
//
// tile: array of 8bit values representing a block of pixels with
//       dimension equal to the 'tilewidth' & 'tileheight' parameters to
//       ::specific_init().  Each value specifies an index into the
//       array of colors you allocated for ::palette_change()
// x0: x origin of tile
// y0: y origin of tile
//
// note: origin of tile and of window based on (0,0) being in the upper
//       left of the window.

void bx_gui_c::graphics_tile_update(Bit8u *tile, unsigned x0, unsigned y0)
{
  if ( !gui_graphics_tile_update )
     {
     bx_printf( "gui dll entry not found for gui_graphics_tile_update\n" );
     return;
     }
  gui_graphics_tile_update(tile,x0,y0);
}



// ::DIMENSION_UPDATE()
//
// Called when the VGA mode changes it's X,Y dimensions.
// Resize the window to this size, but you need to add on
// the height of the headerbar to the Y value.
//
// x: new VGA x size
// y: new VGA y size (add headerbar_y parameter from ::specific_init().

void bx_gui_c::dimension_update(unsigned x, unsigned y)
{
  if ( !gui_dimension_update )
     {
     bx_printf( "gui dll entry not found for gui_dimension_update\n" );
     return;
     }
  gui_dimension_update(x,y);
}


// ::CREATE_BITMAP()
//
// Create a monochrome bitmap of size 'xdim' by 'ydim', which will
// be drawn in the headerbar.  Return an integer ID to the bitmap,
// with which the bitmap can be referenced later.
//
// bmap: packed 8 pixels-per-byte bitmap.  The pixel order is:
//       bit0 is the left most pixel, bit7 is the right most pixel.
// xdim: x dimension of bitmap
// ydim: y dimension of bitmap

unsigned bx_gui_c::create_bitmap(const unsigned char *bmap, unsigned xdim, unsigned ydim)
{
  return(0);
}


// ::HEADERBAR_BITMAP()
//
// Called to install a bitmap in the bochs headerbar (toolbar).
//
// bmap_id: will correspond to an ID returned from
//     ::create_bitmap().  'alignment' is either BX_GRAVITY_LEFT
//     or BX_GRAVITY_RIGHT, meaning install the bitmap in the next
//     available leftmost or rightmost space.
// alignment: is either BX_GRAVITY_LEFT or BX_GRAVITY_RIGHT,
//     meaning install the bitmap in the next
//     available leftmost or rightmost space.
// f: a 'C' function pointer to callback when the mouse is clicked in
//     the boundaries of this bitmap.

unsigned bx_gui_c::headerbar_bitmap(unsigned bmap_id, unsigned alignment, void (*f)(void))
{
  return(0);
}


// ::SHOW_HEADERBAR()
//
// Show (redraw) the current headerbar, which is composed of
// currently installed bitmaps.

void bx_gui_c::show_headerbar(void)
{
}


// ::REPLACE_BITMAP()
//
// Replace the bitmap installed in the headerbar ID slot 'hbar_id',
// with the one specified by 'bmap_id'.  'bmap_id' will have
// been generated by ::create_bitmap().  The old and new bitmap
// must be of the same size.  This allows the bitmap the user
// sees to change, when some action occurs.  For example when
// the user presses on the floppy icon, it then displays
// the ejected status.
//
// hbar_id: headerbar slot ID
// bmap_id: bitmap ID

void bx_gui_c::replace_bitmap(unsigned hbar_id, unsigned bmap_id)
{
}


// ::EXIT()
//
// Called before bochs terminates, to allow for a graceful
// exit from the native GUI mechanism.

void bx_gui_c::exit(void)
{
  if ( !gui_exit )
     {
     bx_printf( "gui dll entry not found for gui_exit\n" );
     return;
     }
  gui_exit();
}
