// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
//
//    US Mail:  Kevin Lawton
//              528 Lexington St.
//              Waltham, MA 02154
//
//    EMail:    bochs@world.std.com

//  Much of this file was written by:
//  Craig S Ballantyne

#define INCL_DOS
#define INCL_DOSDEVICES
#define INCL_DOSDEVIOCTL
#define INCL_VIO
#define INCL_WIN

#include "bochs.h"
#include "../../font/vga.bitmap.h"

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#include "..\..\os2\os2cp.h"

typedef struct
{
   CHAR cBlue;
   CHAR cGreen;
   CHAR cRed;
} myRGB;

#define WM_SET_NEXT_EVENT  WM_USER+1
#define WM_GET_NEXT_EVENT  WM_USER+2
#define WM_CREATE_BITMAP   WM_USER+3
#define WM_DISPLAY_BITMAP  WM_USER+4
#define WM_REPAINT_BITMAP  WM_USER+5
#define WM_CLEAR_SCREEN    WM_USER+6
#define WM_RESIZE_SCREEN   WM_USER+7
#define WM_PALETTE_CHANGE  WM_USER+8
#define WM_TILE_CHANGE     WM_USER+9

bx_options_t *pbx_options;
bx_mem_c *pbx_mem;
static Boolean (*pFloppyAHandler)(Boolean);
static Boolean (*pFloppyBHandler)(Boolean);
static Boolean (*pCDRomHandler)(Boolean);
static Boolean (*pPrinterHandler)(Boolean,UINT);
static void (*pPutScanCode)(PUCHAR,INT);
static void (*pMouseMotion)(INT,INT,UINT);
static HMODULE hModuleID;
static PDLGTEMPLATE pDlgTemplate;

void APIENTRY gui_dll_init( 
   bx_options_t *poptions, 
   bx_mem_c *pmem, 
   Boolean (*pFloppyA)(Boolean), 
   Boolean (*pFloppyB)(Boolean), 
   Boolean (*pCDRom)(Boolean),
   Boolean (*pPrinter)(Boolean,UINT),
   void (*pKey)(PUCHAR,INT),
   void (*pMouse)(INT,INT,UINT),
   HMODULE hModule )
{
   pbx_options     = poptions;
   pbx_mem         = pmem;
   pFloppyAHandler = pFloppyA;
   pFloppyBHandler = pFloppyB;
   pCDRomHandler   = pCDRom;
   pPrinterHandler = pPrinter;
   pPutScanCode    = pKey;
   pMouseMotion    = pMouse;
   hModuleID       = hModule;

  os2Open(0);
}

void gui_put_scancode ( PUCHAR puc, INT i )
{
   pPutScanCode(puc,i);
}

void gui_mouse_motion (INT i1, INT i2, UINT ui)
{
   pMouseMotion(i1,i2,ui);
}

// *********************************************************************
// BOCHS General Functions
// *********************************************************************

void gui_printf(char *fmt, ...)
{
  va_list ap;
  FILE *f;

  va_start(ap, fmt);
  f = fopen( "gui.dbg", "a" );
  vfprintf(f, fmt, ap);
  fflush(f);
  fclose(f);
  va_end(ap);
}

// GUI thread must be dead/done in order to call gui_terminate
void gui_terminate(char *reason)
{
   DosFreeResource( pDlgTemplate );
   os2Close();
   exit(0);
}

// *********************************************************************
// BOCHS Interface Support Functions
// *********************************************************************

static MRESULT APIENTRY dpControlPanel(HWND hwnd, ULONG iMsg, MPARAM mp1, MPARAM mp2)
{
   static BOOL bFloppyA;
   static BOOL bFloppyB;
   static BOOL bCdrom;
   static BOOL bPrinter1;
   static BOOL bPrinter2;
   static BOOL bPrinter3;

   HWND hControl;
   HWND hParent;

   hParent = WinQueryWindow(hwnd,QW_OWNER);

   switch (iMsg)
      {
      case WM_INITDLG:
         bFloppyA = pFloppyAHandler(true);
         bFloppyB = pFloppyBHandler(true);
         bCdrom = pCDRomHandler(true);
         bPrinter1 = pPrinterHandler(true,0);
         bPrinter2 = pPrinterHandler(true,1);
         bPrinter3 = pPrinterHandler(true,2);

         hControl = WinWindowFromID(hwnd,IDE_FLOPPYA_PATH);
         WinSetWindowText(hControl, pbx_options->floppya.path);
         WinEnableWindow(hControl,bFloppyA);
         hControl = WinWindowFromID(hwnd,IDPB_FLOPPYA_BROWSE);
         WinEnableWindow(hControl,bFloppyA);

         hControl = WinWindowFromID(hwnd,IDE_FLOPPYB_PATH);
         WinSetWindowText(hControl, pbx_options->floppyb.path);
         WinEnableWindow(hControl,bFloppyB);
         hControl = WinWindowFromID(hwnd,IDPB_FLOPPYB_BROWSE);
         WinEnableWindow(hControl,bFloppyB);

         hControl = WinWindowFromID(hwnd,IDE_CDROM_PATH);
         WinSetWindowText(hControl, pbx_options->cdromd.dev);
         WinEnableWindow(hControl,bCdrom);
         hControl = WinWindowFromID(hwnd,IDPB_CDROM_BROWSE);
         WinEnableWindow(hControl,bCdrom);

         hControl = WinWindowFromID(hwnd,IDE_PRINTER1_PATH);
         WinSetWindowText(hControl, pbx_options->lpt[0].path);
         WinEnableWindow(hControl,bPrinter1);

         hControl = WinWindowFromID(hwnd,IDE_PRINTER2_PATH);
         WinSetWindowText(hControl, pbx_options->lpt[1].path);
         WinEnableWindow(hControl,bPrinter2);

         hControl = WinWindowFromID(hwnd,IDE_PRINTER3_PATH);
         WinSetWindowText(hControl, pbx_options->lpt[2].path);
         WinEnableWindow(hControl,bPrinter3);

         return (MRESULT) 1;

      case WM_CONTROL:
         return (MRESULT) 1;

      case WM_COMMAND:
         switch(SHORT1FROMMP(mp1))
            {
            case IDPB_STARTMENU:
               // simulate pressing CTRL-ESC
               WinSendMsg( hParent, WM_SET_NEXT_EVENT, MPFROMLONG(0x1d), MPFROMLONG(BX_KEY_PRESSED) );
               WinSendMsg( hParent, WM_SET_NEXT_EVENT, MPFROMLONG(0x01), MPFROMLONG(BX_KEY_PRESSED) );
               WinSendMsg( hParent, WM_SET_NEXT_EVENT, MPFROMLONG(0x01), MPFROMLONG(BX_KEY_RELEASED) );
               WinSendMsg( hParent, WM_SET_NEXT_EVENT, MPFROMLONG(0x1d), MPFROMLONG(BX_KEY_RELEASED) );
               break;

            case IDPB_TASKLIST:
               // simulate pressing ALT-TAB
               WinSendMsg( hParent, WM_SET_NEXT_EVENT, MPFROMLONG(0x38), MPFROMLONG(BX_KEY_PRESSED) );
               WinSendMsg( hParent, WM_SET_NEXT_EVENT, MPFROMLONG(0x0f), MPFROMLONG(BX_KEY_PRESSED) );
               WinSendMsg( hParent, WM_SET_NEXT_EVENT, MPFROMLONG(0x0f), MPFROMLONG(BX_KEY_RELEASED) );
               WinSendMsg( hParent, WM_SET_NEXT_EVENT, MPFROMLONG(0x38), MPFROMLONG(BX_KEY_RELEASED) );
               break;

            case IDPB_CTRLALTDEL:
               // simulate pressing CTRL-ALT-DEL
               WinSendMsg( hParent, WM_SET_NEXT_EVENT, MPFROMLONG(0x1d), MPFROMLONG(BX_KEY_PRESSED) );
               WinSendMsg( hParent, WM_SET_NEXT_EVENT, MPFROMLONG(0x38), MPFROMLONG(BX_KEY_PRESSED) );
               WinSendMsg( hParent, WM_SET_NEXT_EVENT, MPFROMLONG(0x53), MPFROMLONG(BX_KEY_PRESSED) );
               WinSendMsg( hParent, WM_SET_NEXT_EVENT, MPFROMLONG(0x53), MPFROMLONG(BX_KEY_RELEASED) );
               WinSendMsg( hParent, WM_SET_NEXT_EVENT, MPFROMLONG(0x38), MPFROMLONG(BX_KEY_RELEASED) );
               WinSendMsg( hParent, WM_SET_NEXT_EVENT, MPFROMLONG(0x1d), MPFROMLONG(BX_KEY_RELEASED) );
               break;

            case IDPB_POWER:
               if ( WinMessageBox( HWND_DESKTOP, hwnd, "Are you sure you want to quit?", "Power Off", 0, MB_ICONQUESTION | MB_YESNO ) == MBID_YES )
                  WinPostMsg( hParent, WM_CLOSE, 0, 0 );
               break;

            case IDPB_FLOPPYA_BROWSE:
               {
               FILEDLG FileDlg = {0};
               CHAR szDrive[3];

               FileDlg.cbSize = sizeof(FILEDLG);
               FileDlg.fl = FDS_OPEN_DIALOG | FDS_CENTER ;
               FileDlg.pszTitle = (PSZ) "Bochs/2 Open Floppy A:";
               FileDlg.pszOKButton = (PSZ) "Open";
               if ( pbx_options->floppya.path[1] == ':' )
                  {
                  memcpy( (void *) szDrive, (void *) pbx_options->floppya.path, 2 );
                  szDrive[2] = 0;
                  FileDlg.pszIDrive = (PSZ) szDrive;
                  }
               strcpy( FileDlg.szFullFile, pbx_options->floppya.path );
               WinFileDlg( HWND_DESKTOP, hwnd, &FileDlg );
               if ( FileDlg.lReturn != DID_OK )
                  break;
               hControl = WinWindowFromID(hwnd,IDE_FLOPPYA_PATH);
               WinSetWindowText(hControl, FileDlg.szFullFile);
               }
               break;

            case IDPB_FLOPPYA:
               pFloppyAHandler(false);
               bFloppyA = pFloppyAHandler(true);
               hControl = WinWindowFromID(hwnd,IDE_FLOPPYA_PATH);
               WinEnableWindow(hControl,bFloppyA);
               WinQueryWindowText(hControl, sizeof(pbx_options->floppya.path), pbx_options->floppya.path);
               hControl = WinWindowFromID(hwnd,IDPB_FLOPPYA_BROWSE);
               WinEnableWindow(hControl,bFloppyA);
               break;

            case IDPB_FLOPPYB_BROWSE:
               {
               FILEDLG FileDlg = {0};
               CHAR szDrive[3];

               FileDlg.cbSize = sizeof(FILEDLG);
               FileDlg.fl = FDS_OPEN_DIALOG | FDS_CENTER ;
               FileDlg.pszTitle = (PSZ) "Bochs/2 Open Floppy B:";
               FileDlg.pszOKButton = (PSZ) "Open";
               if ( pbx_options->floppyb.path[1] == ':' )
                  {
                  memcpy( (void *) szDrive, (void *) pbx_options->floppyb.path, 2 );
                  szDrive[2] = 0;
                  FileDlg.pszIDrive = (PSZ) szDrive;
                  }
               strcpy( FileDlg.szFullFile, pbx_options->floppyb.path );
               WinFileDlg( HWND_DESKTOP, hwnd, &FileDlg );
               if ( FileDlg.lReturn != DID_OK )
                  break;
               hControl = WinWindowFromID(hwnd,IDE_FLOPPYB_PATH);
               WinSetWindowText(hControl, FileDlg.szFullFile);
               }
               break;

            case IDPB_FLOPPYB:
               pFloppyBHandler(false);
               bFloppyB = pFloppyBHandler(true);
               hControl = WinWindowFromID(hwnd,IDE_FLOPPYB_PATH);
               WinEnableWindow(hControl,bFloppyB);
               WinQueryWindowText(hControl, sizeof(pbx_options->floppyb.path), pbx_options->floppyb.path);
               hControl = WinWindowFromID(hwnd,IDPB_FLOPPYB_BROWSE);
               WinEnableWindow(hControl,bFloppyB);
               break;

            case IDPB_CDROM:
               pCDRomHandler(false);
               bCdrom = pCDRomHandler(true);
               hControl = WinWindowFromID(hwnd,IDE_CDROM_PATH);
               WinEnableWindow(hControl,bCdrom);
               WinQueryWindowText(hControl, sizeof(pbx_options->cdromd.dev), pbx_options->cdromd.dev);
               hControl = WinWindowFromID(hwnd,IDPB_CDROM_BROWSE);
               WinEnableWindow(hControl,bCdrom);
               break;

            case IDPB_PRINTER1:
               pPrinterHandler(false,0);
               bPrinter1 = pPrinterHandler(true,0);
               hControl = WinWindowFromID(hwnd,IDE_PRINTER1_PATH);
               WinEnableWindow(hControl,bPrinter1);
               WinQueryWindowText(hControl, sizeof(pbx_options->lpt[0].path), pbx_options->lpt[0].path);
               break;

            case IDPB_PRINTER2:
               pPrinterHandler(false,0);
               bPrinter2 = pPrinterHandler(true,1);
               hControl = WinWindowFromID(hwnd,IDE_PRINTER2_PATH);
               WinEnableWindow(hControl,bPrinter2);
               WinQueryWindowText(hControl, sizeof(pbx_options->lpt[1].path), pbx_options->lpt[1].path);
               break;

            case IDPB_PRINTER3:
               pPrinterHandler(false,0);
               bPrinter3 = pPrinterHandler(true,2);
               hControl = WinWindowFromID(hwnd,IDE_PRINTER3_PATH);
               WinEnableWindow(hControl,bPrinter3);
               WinQueryWindowText(hControl, sizeof(pbx_options->lpt[2].path), pbx_options->lpt[2].path);
               break;
            }

         return (MRESULT) 1;
      }

   return WinDefDlgProc (hwnd, iMsg, mp1, mp2);
}

HWND gui_control_panel( HWND hWindow )
{
   HWND hDialog;
   ERRORID eid;

   hDialog = WinLoadDlg( HWND_DESKTOP, hWindow, dpControlPanel, 0, DLG_BOCHS_CONTROLPANEL, 0 );

   eid = WinGetLastError(0);

   return hDialog;
}

