// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
//
//    US Mail:  Kevin Lawton
//              528 Lexington St.
//              Waltham, MA 02154
//
//    EMail:    bochs@world.std.com

//  Much of this file was written by:
//  Craig S Ballantyne

#define INCL_DOS
#define INCL_DOSSEMAPHORES
#define INCL_DOSPROCESS
#define INCL_DOSMISC
#define INCL_GPI
#define INCL_WIN
#define INCL_VIO
#define INCL_GRE_COLORTABLE
#define INCL_GRE_DEVMISC3
#include "bochs.h"

#include "../../font/vga.bitmap.h"

#define  _MEERROR_H_
#include <mmioos2.h>                    /* It is from MMPM toolkit           */
#include <dive.h>
#include <fourcc.h>
//#include <builtin.h>
#include <pmddim.h>

#include "..\..\os2\os2cp.h"

#include <process.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#define WM_SET_NEXT_EVENT     WM_USER+1
#define WM_GET_NEXT_EVENT     WM_USER+2

#define TILESIZE_X   16
#define TILESIZE_Y   16
#define VGA_X        640
#define VGA_Y        480

#ifdef __EMX__
#define min(a, b)       (((a) < (b))? (a) : (b))
#define max(a, b)       (((a) > (b))? (a) : (b))
#endif

// Keyboard/mouse stuff
#define SCANCODE_BUFSIZE    100
#define LMOUSE_PRESSED      0x20000000
#define RMOUSE_PRESSED      0x40000000
#define LMOUSE_RELEASED     0x02000000
#define RMOUSE_RELEASED     0x04000000

static Bit32u keyevents[SCANCODE_BUFSIZE];
static unsigned screen_updated=0;
static unsigned head=0, tail=0;
static unsigned mouse_button_state = 0;
static unsigned ms_curr_x=0, ms_curr_y=0, ms_prev_x=0, ms_prev_y=0;
static int workerThread = 0;

// mutex semaphore
static HMTX hmtx = 0;

// DIVE stuff
static HDIVE             hDive;
static SETUP_BLITTER     SetupBlitter;
static ULONG             ulImage;
static BOOL              fVrnDisabled;
static ULONG             ulColors;
static DIVE_CAPS         DiveCaps = {0};
static FOURCC            fccFormats[100] = {0};
static RECTL             rclUpdate;

// Graphics screen stuff
static RGB2              rgbPalette[256];
static unsigned char    *vgaChar256[256];
static unsigned int      usPaletteSize;

// Misc stuff
static unsigned dimension_x, dimension_y;
static unsigned window_x, window_y;
static unsigned prev_block_cursor_x = 0;
static unsigned prev_block_cursor_y = 0;
static unsigned y_edge=0, x_edge=0, y_caption=0;

static char *pszAppName = "Bochs/2";
static char *pszWindowName = "Bochs/2 VGA Monitor";

typedef struct
{
   char    *kill;  // reason for terminateEmul(int)
   BOOL     UIinited;
   HWND     hwnd;
   HWND     frame;
   HWND     dialog;
} sharedThreadInfo;

static sharedThreadInfo stInfo;

// *********************************************************************
// BOCHS General Functions
// *********************************************************************

extern void gui_put_scancode(PUCHAR puc, INT i);
extern void gui_mouse_motion(INT i1, INT i2, UINT ui);
extern HWND gui_control_panel(HWND hwnd);
extern void gui_terminate(char *reason);

static void diveDisplay( PRECTL prcl )
{
   PBYTE    pbBuffer;
   PBYTE    pSrcAddress;
   PBYTE    pDestAddress;         /* destination VRAM adrress               */
   ULONG    ulScanLineBytes;
   ULONG    ulScanLines;
   BOOL     fFirstLineInRect;     /* Indicates start of drawing sequence    */
   RECTL    rcl;                  /* Current rectangle to draw              */
   ULONG    i;
   ULONG    ulCopyLengthBytes;    /* Length of scan line for current window */
   ULONG    ulMoreLine;           /* Number of lines left to draw           */
   ULONG    ulColorSizeBytes;
   ULONG    ulLineSizeBytes;
   ULONG    ulRemLinesInBank;     /* Number of lines left in bank           */
   ULONG    ulBankNumber;         /* Bank number returned from DIVE         */
   ULONG    ulTmpImage;
   ULONG    ulRC;

   if ( !ulImage || fVrnDisabled )
      return;

   DosRequestMutexSem( hmtx, (ULONG) SEM_INDEFINITE_WAIT );

   ulRC = DiveSetSourcePalette ( hDive, 0, usPaletteSize, &rgbPalette[0].bBlue );

   ulRC = DiveBlitImage( hDive, ulImage, 0 );
   return;

#ifdef UNBROKEN_BY_SCITECH

   if( !DiveCaps.fScreenDirect || ( DiveCaps.fBankSwitched && DiveCaps.ulApertureSize % DiveCaps.ulScanLineBytes ) )
       {
       ulRC = DiveBlitImage( hDive, ulImage, 0 );
       return;
       }

   ulTmpImage = 0;
   ulRC = DiveAllocImageBuffer( hDive, &ulTmpImage, FOURCC_SCRN, dimension_x, dimension_y+2, 0, 0 );

   ulRC = DiveBlitImage( hDive, ulImage, ulTmpImage );

   pbBuffer = 0;
   ulScanLineBytes = 0;
   ulScanLines = 0;
   ulRC = DiveBeginImageBufferAccess( hDive, ulTmpImage, &pbBuffer, &ulScanLineBytes, &ulScanLines );

   ulColorSizeBytes = DiveCaps.ulScanLineBytes / DiveCaps.ulHorizontalResolution ;

   ulLineSizeBytes = (dimension_x * ulColorSizeBytes)/4*4;

   if (dimension_x * ulColorSizeBytes - ulLineSizeBytes)
      ulLineSizeBytes += 4;

   for (i = 0; i < SetupBlitter.ulNumDstRects; i++)
      {
      pSrcAddress = pbBuffer;

      memcpy (&rcl, &SetupBlitter.pVisDstRects[i], sizeof(RECTL) );

      // Size of scan line bounded by this rectangle
      ulCopyLengthBytes = ( rcl.xRight - rcl.xLeft ) * ulColorSizeBytes;

      // Number of lines contained in this rectangle
      ulMoreLine = rcl.yTop - rcl.yBottom;

      // Start position in source bitmap
      pSrcAddress += ( dimension_y - rcl.yTop ) *
                     ulLineSizeBytes +
                     ( rcl.xLeft * ulColorSizeBytes );

      // Map rectangle to desktop coordinates
      rcl.yTop    += SetupBlitter.lScreenPosY;
      rcl.yBottom += SetupBlitter.lScreenPosY;
      rcl.xRight  += SetupBlitter.lScreenPosX;
      rcl.xLeft   += SetupBlitter.lScreenPosX;

      /* NOTICE !!! */
      /**** TEMPORARY ADJUSTMENT  start ****************************
      * If this adjustment is not here, DiveCalcFrameBufferAddress
      * returns 1 dot higher position.
      * Need to contact the devloper.
      *************************************************************/
      rcl.yTop --;
      rcl.yBottom --;
      /* end temporary adjustment */

      fFirstLineInRect = TRUE;

      while ( ulMoreLine )
         {
         if ( fFirstLineInRect )
            {
            // Get VRAM start location
            pDestAddress = 0;
            ulBankNumber = 0;
            ulRemLinesInBank = 0;
            ulRC = DiveCalcFrameBufferAddress ( hDive, &rcl, &pDestAddress, &ulBankNumber, &ulRemLinesInBank );

            if ( DiveAcquireFrameBuffer ( hDive, &rcl ) )
               {
               break;
               }

            ulRC = DiveSwitchBank ( hDive, ulBankNumber );

            fFirstLineInRect = FALSE;
            }

         // Copy one Line to display
         memcpy ( pDestAddress, pSrcAddress, ulCopyLengthBytes );

         // Update source address
         pSrcAddress += ulLineSizeBytes;

         ulMoreLine--;
         ulRemLinesInBank--;
         rcl.yTop--;

         // Check if time to switch banks.
         if ( !ulRemLinesInBank )
            {
            ulRC = DiveCalcFrameBufferAddress ( hDive, &rcl, &pDestAddress, &ulBankNumber, &ulRemLinesInBank );

            ulRC = DiveSwitchBank ( hDive, ulBankNumber );
            }
         else
            {
            // update destination address.
            pDestAddress += DiveCaps.ulScanLineBytes;
            }
         }

      ulRC = DiveDeacquireFrameBuffer ( hDive );
      }

   ulRC = DiveEndImageBufferAccess( hDive, ulTmpImage );

   ulRC = DiveFreeImageBuffer( hDive, ulTmpImage );

#endif

   DosReleaseMutexSem( hmtx );
}

// *********************************************************************
// Display Window Functions
// *********************************************************************

static MRESULT APIENTRY wpBochsDisplay(HWND hwnd, ULONG iMsg, MPARAM mp1, MPARAM mp2)
{
   static RECTL      rcls[50];

   HPS               hps;
   HRGN              hrgn;
   RGNRECT           rgnCtl;
   Bit32u            key;
   Bit32u            press_release;
   int               cx, cy;
   int               i, j, k;
   SWP               swp;
   POINTL            pointl;
   unsigned char    *data256;
   int               c;
   RECTL             rcl;


   switch (iMsg)
      {
      case WM_CREATE:
         cx       = 8;
         cy       = 16;

         for ( c = 0 ; c < 256 ; c++ )
            {
            unsigned char data[16];

            for (i=0; i<16; i++)
               data[i] = ~bx_vgafont[c].data[i];

            data256  = (unsigned char *) calloc( cx, cy );

            for ( i = 0 ; i < cy ; i++ )
               {
               for ( j = 0 ; j < cx/8 ; j++ )
                  {
                  char c = data[(cy-i-1)*cx/8+j];
   
                  for ( k = 0 ; k < 8 ; k++ )
                     {
                     if ( c & 1 )
                        data256[(cy-i-1)*cx+j*8+k] = 0 ;
                     else
                        data256[(cy-i-1)*cx+j*8+k] = 7 ;

                     c >>= 1;
                     }
                  }
               }

            vgaChar256[c] = data256;
            }

          stInfo.dialog = gui_control_panel( hwnd );

          break;

      case WM_VRNDISABLED:
          DosRequestMutexSem( hmtx, (ULONG) SEM_INDEFINITE_WAIT );

          fVrnDisabled = TRUE;
          DiveSetupBlitter ( hDive, 0 );

          DosReleaseMutexSem( hmtx );
          break;

      case WM_VRNENABLED:
          DosRequestMutexSem( hmtx, (ULONG) SEM_INDEFINITE_WAIT );

          hps  = WinGetPS ( hwnd );
          hrgn = GpiCreateRegion ( hps, 0L, NULL );

          WinQueryVisibleRegion ( hwnd, hrgn );

          rgnCtl.ircStart     = 1;
          rgnCtl.crc          = 50;
          rgnCtl.ulDirection  = 1;

          // Get the all ORed rectangles

          GpiQueryRegionRects ( hps, hrgn, NULL, &rgnCtl, rcls);

          GpiDestroyRegion( hps, hrgn );

          WinReleasePS( hps );

          // Now find the window position and size, relative to parent.

          WinQueryWindowPos ( hwnd, &swp );

          // Convert the point to offset from desktop lower left.

          pointl.x = swp.x;
          pointl.y = swp.y;

          WinMapWindowPoints ( stInfo.frame, HWND_DESKTOP, &pointl, 1 );

          // Tell DIVE about the new settings.

          SetupBlitter.ulStructLen       = sizeof( SETUP_BLITTER );
          SetupBlitter.fccSrcColorFormat = FOURCC_LUT8;
          SetupBlitter.ulSrcWidth        = dimension_x;
          SetupBlitter.ulSrcHeight       = dimension_y;
          SetupBlitter.ulSrcPosX         = 0;
          SetupBlitter.ulSrcPosY         = 0;
          SetupBlitter.fInvert           = TRUE;
          SetupBlitter.ulDitherType      = 1;

          SetupBlitter.fccDstColorFormat = FOURCC_SCRN;
          SetupBlitter.ulDstWidth        = dimension_x;
          SetupBlitter.ulDstHeight       = dimension_y;
          SetupBlitter.lDstPosX          = 0;
          SetupBlitter.lDstPosY          = 0;
          SetupBlitter.lScreenPosX       = pointl.x;
          SetupBlitter.lScreenPosY       = pointl.y;
          SetupBlitter.ulNumDstRects     = rgnCtl.crcReturned;
          SetupBlitter.pVisDstRects      = rcls;

          DiveSetupBlitter ( hDive, &SetupBlitter );
          fVrnDisabled = FALSE;

          DosReleaseMutexSem( hmtx );

          stInfo.UIinited = TRUE;

          break;

      case WM_REALIZEPALETTE:
          // This tells DIVE that the physical palette may have changed.
          DiveSetDestinationPalette ( hDive, 0, 0, 0 );
          break;

      case WM_CLOSE:
         WinDismissDlg( stInfo.dialog, 0 );
         WinPostMsg( hwnd, WM_QUIT, 0, 0 );
         break;

      case WM_MOUSEMOVE:
         ms_curr_x = SHORT1FROMMP(mp1);
         ms_curr_y = SHORT2FROMMP(mp1);
         break;

      case WM_BUTTON1DOWN:
         mouse_button_state |= 0x01;
         WinSendMsg( hwnd, WM_SET_NEXT_EVENT, MPFROMLONG(mouse_button_state), MPFROMLONG(LMOUSE_PRESSED) );
         break;
    
      case WM_BUTTON1UP:
         mouse_button_state &= ~0x01;
         WinSendMsg( hwnd, WM_SET_NEXT_EVENT, MPFROMLONG(mouse_button_state), MPFROMLONG(LMOUSE_RELEASED) );
         break;

      case WM_BUTTON2DOWN:
         mouse_button_state |= 0x02;
         WinSendMsg( hwnd, WM_SET_NEXT_EVENT, MPFROMLONG(mouse_button_state), MPFROMLONG(RMOUSE_PRESSED) );
         break;
    
      case WM_BUTTON2UP:
         mouse_button_state &= ~0x02;
         WinSendMsg( hwnd, WM_SET_NEXT_EVENT, MPFROMLONG(mouse_button_state), MPFROMLONG(RMOUSE_RELEASED) );
         break;
    
      case WM_CHAR:
         if ( SHORT1FROMMP(mp1) & KC_SCANCODE )
            {
            if ( SHORT1FROMMP(mp1) & KC_KEYUP )
               {
               WinSendMsg( hwnd, WM_SET_NEXT_EVENT, MPFROMLONG(CHAR4FROMMP(mp1)), MPFROMLONG(BX_KEY_RELEASED) );
               }
            else
               {
               WinSendMsg( hwnd, WM_SET_NEXT_EVENT, MPFROMLONG(CHAR4FROMMP(mp1)), MPFROMLONG(BX_KEY_PRESSED) );
               }
            }
         return (MRESULT) 1;

      case WM_SET_NEXT_EVENT:
         key = LONGFROMMP(mp1);
         press_release = LONGFROMMP(mp2);

         if (((tail+1) % SCANCODE_BUFSIZE) == head)
            {
            fprintf(stderr, "# enq_scancode: buffer full\n");
            }
         else
            {
            keyevents[tail] = key | press_release;
            tail = (tail + 1) % SCANCODE_BUFSIZE;
            }
         break;

      case WM_GET_NEXT_EVENT:
         if ( head == tail )
            {
            fprintf(stderr, "# deq_scancode: buffer empty\n");
            return(0);
            }
         else
            {
            key = keyevents[head];
            head = (head + 1) % SCANCODE_BUFSIZE;
            return (MRESULT)key;
            }
         break;

      case WM_PAINT:
         hps = WinBeginPaint( hwnd, NULL, &rcl );
         diveDisplay( &rcl );
         WinEndPaint( hps );
         break;
   }

   return WinDefWindowProc (hwnd, iMsg, mp1, mp2);
}

// This thread controls the GUI window.
static VOID _Optlink UIThread(PVOID pvoid)
{
   HAB   hab;
   HMQ   hmq;
   QMSG  qmsg;
   ULONG flCreate;

   fVrnDisabled = TRUE;

   // Get the screen capabilities, and if the system support only 16 colors
   // the sample should be terminated.
   DiveCaps.pFormatData = fccFormats;
   DiveCaps.ulFormatLength = 120;
   DiveCaps.ulStructLen = sizeof(DIVE_CAPS);

   if ( DiveQueryCaps ( &DiveCaps, DIVE_BUFFER_SCREEN ) ||
        DiveCaps.ulDepth < 8 ||
        DiveOpen( &hDive, FALSE, 0 ) )
     {
     WinMessageBox( HWND_DESKTOP, HWND_DESKTOP,
              "This program cannot run on this system.",
              pszWindowName, 0, MB_OK | MB_INFORMATION );
     _endthread();
     }

   DosCreateMutexSem( "\\sem32\bochs2", &hmtx,0 , FALSE );

   hab = WinInitialize(0);
   hmq = WinCreateMsgQueue(hab,0);

   WinRegisterClass(hab,pszAppName,wpBochsDisplay,0,sizeof(LONG));

   flCreate = FCF_TASKLIST | FCF_TITLEBAR | FCF_MINBUTTON | FCF_DLGBORDER ;

   stInfo.frame = WinCreateStdWindow(HWND_DESKTOP,
      WS_VISIBLE,
      &flCreate,
      pszAppName,
      pszWindowName,
      WS_VISIBLE,
      0,
      666,
      &stInfo.hwnd );

   if (!stInfo.frame) gui_terminate("window creation error!\n");

   WinSetFocus( HWND_DESKTOP, stInfo.hwnd );

   dimension_x = VGA_X;
   dimension_y = VGA_Y;
  
   x_edge      = WinQuerySysValue(HWND_DESKTOP,SV_CXDLGFRAME);
   y_edge      = WinQuerySysValue(HWND_DESKTOP,SV_CYDLGFRAME);
   y_caption   = WinQuerySysValue(HWND_DESKTOP,SV_CYTITLEBAR);

   window_x    = 2*x_edge;
   window_y    = y_caption + 2*y_edge;

   WinSetVisibleRegionNotify( stInfo.hwnd, TRUE );

   WinPostMsg( stInfo.frame, WM_VRNENABLED, 0, 0 );

   WinSetWindowPos(stInfo.frame,
      HWND_TOP,
      0,
      0,
      0,
      0,
      SWP_SHOW | SWP_ACTIVATE );

   while (WinGetMsg (hab, &qmsg, 0, 0, 0))
      WinDispatchMsg (hab, &qmsg);

   WinSetVisibleRegionNotify( stInfo.hwnd, FALSE );

   WinDestroyWindow( stInfo.frame );
   WinDestroyMsgQueue( hmq );
   WinTerminate( hab );

   DosCloseMutexSem( hmtx );

   DiveFreeImageBuffer( hDive, ulImage );
   DiveClose( hDive );

   hDive = 0;

   stInfo.kill = "Window closed, exiting!\n";

   _endthread();
}

// *********************************************************************
// BOCHS Interface Functions
// *********************************************************************

void APIENTRY gui_specific_init(bx_gui_c *th, int argc, char **argv, unsigned tilewidth, unsigned tileheight, unsigned headerbar_y)
{
   HAB   hab;
   HMQ   hmq;

   hab = WinInitialize(0);
   if (!hab) gui_terminate("anchor block creation error!\n");

   hmq = WinCreateMsgQueue(hab,0);
   if (!hmq) gui_terminate("message queue creation error!\n");

   stInfo.kill = 0;
   stInfo.UIinited = FALSE;

   workerThread = _beginthread (UIThread, 0, 32768, 0);

   if (workerThread == -1) gui_terminate("thread creation error!\n");
  
   // Wait for a window before continuing
   while (!stInfo.UIinited) DosSleep(10);
  
   // Now set this thread's priority to below normal because this is where
   // the emulated CPU runs, and it hogs the real CPU
   DosSetPriority(PRTYS_THREAD, PRTYC_IDLETIME, 0, 0);
}

void APIENTRY gui_handle_events(void)
{
   Bit32u key;
   unsigned char scancode;
   unsigned ms_xdelta = ms_curr_x - ms_prev_x;
   unsigned ms_ydelta = ms_curr_y - ms_prev_y;
  
   // printf("# Hey!!!\n");

   if (stInfo.kill) gui_terminate(stInfo.kill);
  
   if ( screen_updated )
      {
      usPaletteSize = 256;
      diveDisplay( &rclUpdate );
      screen_updated = 0;
      }

   ms_prev_x = ms_curr_x;
   ms_prev_y = ms_curr_y;

   // Handle mouse moves
   if (ms_xdelta || ms_ydelta)
      {
      gui_mouse_motion(ms_xdelta, ms_ydelta, mouse_button_state);
      }
  
  // Handle keyboard and mouse clicks
   while (head != tail)
      {
      // Bypass bx_devices.keyboard->gen_scancode so we may enter
      // a scancode directly bx_devices.keyboard->gen_scancode();
      key = (Bit32u) WinSendMsg( stInfo.hwnd, WM_GET_NEXT_EVENT, 0, 0 );

      // Check for mouse buttons first
      if ((key & LMOUSE_PRESSED) || (key & LMOUSE_RELEASED) || (key & RMOUSE_PRESSED) || (key & RMOUSE_RELEASED))
         {
         // printf("# click!\n");
         gui_mouse_motion( 1, 1, mouse_button_state);
         gui_mouse_motion( 0, 0, key);
         gui_mouse_motion( -1, -1, mouse_button_state);
         }
      else
         {
         // Its a key
         // printf("# key = %d, scancode = %d\n",key,scancode);
         if (key & BX_KEY_RELEASED)
            {
            scancode = key - BX_KEY_RELEASED;
            scancode |= 0x80;
            }
         else
            {
            scancode = key;
            }
         gui_put_scancode(&scancode, 1);
         }
      }
}

static void DisplayFont(unsigned c, unsigned a, unsigned x, unsigned y, BOOL bNormal)
{
   unsigned char        data256[8*16];
   unsigned             i, j, cx, cy;
   unsigned char        fg, bg;
   PBYTE                pbBuffer;
   ULONG                ulScanLineBytes, ulScanLines;


   if ( !ulImage )
      return;

   memset( data256, 0, 8*16 );

   cx = 8;
   cy = 16;
   x *= cx;
   y *= cy;

   fg = ( a & 0x0f );
   bg = ( ( a >> 4 ) & 0x0f );

   for ( i = 0 ; i < 8 * 16 ; i++ )
      {
      unsigned char x = vgaChar256[c][i] ;

      if ( bNormal )
         data256[i] = x ? fg : bg ;
      else
         data256[i] = x ? bg : fg ;
      }

   DosRequestMutexSem( hmtx, (ULONG) SEM_INDEFINITE_WAIT );

   DiveBeginImageBufferAccess( hDive, ulImage, &pbBuffer, &ulScanLineBytes, &ulScanLines );

   for ( i = 0 ; i < cy ; i++ )
      {
      j = ( dimension_y - y - i - 1 ) * dimension_x + x;
      memcpy( &pbBuffer[j], &data256[i*cx], cx );
      }

   DiveEndImageBufferAccess ( hDive, ulImage );

   DosReleaseMutexSem( hmtx );

   if ( rclUpdate.xLeft == 0 && rclUpdate.xRight == 0 )
      {
      rclUpdate.xLeft   = x;
      rclUpdate.xRight  = x+cx;
      rclUpdate.yTop    = y+cy;
      rclUpdate.yBottom = y;
      }
   else
      {
      if ( rclUpdate.xLeft < x )
         {
         rclUpdate.xLeft   = x;
         rclUpdate.xRight  = x+cx;
         }
      else if ( rclUpdate.yTop > y )
         {
         rclUpdate.yTop    = y+cy;
         rclUpdate.yBottom = y;
         }
      }
}

void APIENTRY gui_text_update(Bit8u *old_text, Bit8u *new_text, unsigned long cursor_x, unsigned long cursor_y, unsigned nrows)
{
   unsigned char achar, aattr;
   unsigned i, x, y;
   unsigned nchars;


   if ( !hDive )
      return;

   usPaletteSize = 8;

   // BLACK
   rgbPalette[0].bBlue     = 0x00;
   rgbPalette[0].bGreen    = 0x00;
   rgbPalette[0].bRed      = 0x00;
   rgbPalette[0].fcOptions = 0;

   // BLUE
   rgbPalette[1].bBlue     = 0xff;
   rgbPalette[1].bGreen    = 0x00;
   rgbPalette[1].bRed      = 0x00;
   rgbPalette[1].fcOptions = 0;

   // GREEN
   rgbPalette[2].bBlue     = 0x00;
   rgbPalette[2].bGreen    = 0xff;
   rgbPalette[2].bRed      = 0x00;
   rgbPalette[2].fcOptions = 0;

   // CYAN
   rgbPalette[3].bBlue     = 0xff;
   rgbPalette[3].bGreen    = 0xff;
   rgbPalette[3].bRed      = 0x00;
   rgbPalette[3].fcOptions = 0;

   // RED
   rgbPalette[4].bBlue     = 0x00;
   rgbPalette[4].bGreen    = 0x00;
   rgbPalette[4].bRed      = 0xff;
   rgbPalette[4].fcOptions = 0;

   // PINK
   rgbPalette[5].bBlue     = 0xff;
   rgbPalette[5].bGreen    = 0x00;
   rgbPalette[5].bRed      = 0xff;
   rgbPalette[5].fcOptions = 0;

   // YELLOW
   rgbPalette[6].bBlue     = 0x00;
   rgbPalette[6].bGreen    = 0xff;
   rgbPalette[6].bRed      = 0xff;
   rgbPalette[6].fcOptions = 0;

   // WHITE
   rgbPalette[7].bBlue     = 0xff;
   rgbPalette[7].bGreen    = 0xff;
   rgbPalette[7].bRed      = 0xff;
   rgbPalette[7].fcOptions = 0;

   nchars = 80 * nrows;

   i = prev_block_cursor_y*80 + prev_block_cursor_x;
   if ( i < nchars )
      {
      DisplayFont(new_text[i*2],new_text[i*2+1], prev_block_cursor_x, prev_block_cursor_y, TRUE);
      }
  
   for ( i = 0; i < nchars * 2; i += 2 )
      {
      if ((old_text[i] != new_text[i]) || (old_text[i+1] != new_text[i+1]))
         {
         x = (i/2) % 80;
         y = (i/2) / 80;

         DisplayFont(new_text[i], new_text[i+1], x, y, TRUE);
         }
      }
  
   prev_block_cursor_x = cursor_x;
   prev_block_cursor_y = cursor_y;
  
   // now draw character at new block cursor location in reverse
   i = cursor_y*80 + cursor_x;
   if ( i < nchars )
      {
      DisplayFont(new_text[i*2], new_text[i*2+1], cursor_x, cursor_y, FALSE);
      }

   diveDisplay( &rclUpdate );
}

Boolean APIENTRY gui_palette_change(unsigned index, unsigned red, unsigned green, unsigned blue)
{
   RGB2     rgb;

   rgb.bBlue     = blue;
   rgb.bGreen    = green;
   rgb.bRed      = red;
   rgb.fcOptions = 0;

   if ( rgbPalette[index].bBlue  == rgb.bBlue &&
        rgbPalette[index].bGreen == rgb.bGreen &&
        rgbPalette[index].bRed   == rgb.bRed )
      return 0;

   rgbPalette[index].bBlue     = rgb.bBlue;
   rgbPalette[index].bGreen    = rgb.bGreen;
   rgbPalette[index].bRed      = rgb.bRed;
   rgbPalette[index].fcOptions = 0;

   if ( index == 0 )
      ulColors = 0;

   ulColors = min( index, ulColors+1 );

   return 0;
}

void APIENTRY gui_graphics_tile_update(Bit8u *tile, unsigned x, unsigned y)
{
   unsigned             i, j;
   PBYTE                pbBuffer;
   ULONG                ulScanLineBytes, ulScanLines;


   if ( !hDive )
      return;

   DosRequestMutexSem( hmtx, (ULONG) SEM_INDEFINITE_WAIT );

   DiveBeginImageBufferAccess( hDive, ulImage, &pbBuffer, &ulScanLineBytes, &ulScanLines );

   for ( i = 0 ; i < TILESIZE_Y ; i++ )
      {
      j = ( dimension_y - y - i - 1 ) * dimension_x + x;
      memcpy( &pbBuffer[j], &tile[i*TILESIZE_X], TILESIZE_X );
      }

   DiveEndImageBufferAccess ( hDive, ulImage );

   DosReleaseMutexSem( hmtx );

   if ( rclUpdate.xLeft == 0 && rclUpdate.xRight == 0 )
      {
      rclUpdate.xLeft   = x;
      rclUpdate.xRight  = x+8;
      rclUpdate.yTop    = y+16;
      rclUpdate.yBottom = y;
      }
   else
      {
      if ( rclUpdate.xLeft < x )
         {
         rclUpdate.xLeft   = x;
         rclUpdate.xRight  = x+8;
         }
      else if ( rclUpdate.yTop > y )
         {
         rclUpdate.yTop    = y+16;
         rclUpdate.yBottom = y;
         }
      }

   screen_updated = 1;
}

void APIENTRY gui_dimension_update(unsigned x, unsigned y)
{
   static BOOL bFirstTime = TRUE ;

   BOOL  bRedraw;

   long pos_x;
   long pos_y;
   long cy_desktop;
   long cx_desktop;
   RECTL    rclDialog = {0};


   if ( !hDive )
      return;

   bRedraw = ( dimension_x != x || dimension_y != y || bFirstTime );

   if ( bRedraw )
      {
      DosRequestMutexSem( hmtx, (ULONG) SEM_INDEFINITE_WAIT );

      if ( ulImage )
         {
         DiveFreeImageBuffer( hDive, ulImage );
         ulImage = 0;
         }

      dimension_x = x;
      dimension_y = y;

      window_x = dimension_x + 2*x_edge;
      window_y = dimension_y + y_caption + 2*y_edge;

      if ( stInfo.dialog )
         WinQueryWindowRect ( stInfo.dialog, &rclDialog );

      cx_desktop = WinQuerySysValue(HWND_DESKTOP,SV_CXSCREEN);
      cy_desktop = WinQuerySysValue(HWND_DESKTOP,SV_CYSCREEN);

      pos_x  = (cx_desktop-window_x)/2;

      if ( cy_desktop > window_y )
         {
         pos_y  = (cy_desktop-window_y)/2 + (rclDialog.yTop - rclDialog.yBottom)/2;
         WinSetWindowPos(stInfo.frame,HWND_TOP,pos_x,pos_y,window_x,window_y,SWP_SIZE | SWP_MOVE );

         pos_y  = (cy_desktop-window_y)/2 - (rclDialog.yTop - rclDialog.yBottom)/2;
         WinSetWindowPos(stInfo.dialog,HWND_TOP,pos_x,pos_y,0,0,SWP_MOVE );
         }
      else
         {
         pos_y  = cy_desktop - window_y;
         WinSetWindowPos(stInfo.frame,HWND_TOP,pos_x,pos_y,window_x,window_y,SWP_SIZE | SWP_MOVE );

         pos_y  = 0;
         WinSetWindowPos(stInfo.dialog,HWND_TOP,pos_x,pos_y,0,0,SWP_MOVE );
         }

      DiveAllocImageBuffer( hDive, &ulImage, FOURCC_LUT8, dimension_x, dimension_y, 0, 0 ) ;

      DosReleaseMutexSem( hmtx );
      bFirstTime = FALSE ;
      }

   rclUpdate.xLeft   = 0;
   rclUpdate.xRight  = 0;
   rclUpdate.yTop    = 0;
   rclUpdate.yBottom = 0;
}

void APIENTRY gui_exit(void)
{
   printf("# In bx_gui_c::exit(void)!\n");
  
   // kill thread first...
   WinPostMsg(stInfo.hwnd, WM_CLOSE, 0, 0);

   // Wait until it dies
   while ((stInfo.kill == 0)) DosSleep(10);

   gui_terminate(NULL);
}

