// Copyright Notice
// ================
// BOCHS is Copyright 1994-1999 by Kevin P. Lawton.
//
// BOCHS is commercial software.
//
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
//
//    US Mail:  Kevin Lawton
//              528 Lexington St.
//              Waltham, MA 02452-3028
//
//    EMail:    bochs@world.std.com



#define INCL_VIO
#define INCL_DOS
#define INCL_KBD
#define INCL_MOU
#include "bochs.h"

#include <process.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

extern void gui_put_scancode ( PUCHAR puc, INT i );
extern void gui_mouse_motion (INT i1, INT i2, UINT ui);
extern void gui_printf(char *fmt, ...);
#define DEBUGLINE gui_printf("LINE=%d\n",__LINE__);

extern bx_options_t *pbx_options;
extern bx_mem_c *pbx_mem;

// Keyboard/mouse stuff
#define SCANCODE_BUFSIZE    256
#define LMOUSE_PRESSED      0x20000000
#define RMOUSE_PRESSED      0x40000000
#define LMOUSE_RELEASED     0x02000000
#define RMOUSE_RELEASED     0x04000000

static Bit32u keyevents[SCANCODE_BUFSIZE];
static unsigned screen_updated=0;
static unsigned head=0, tail=0;
static unsigned mouse_button_state = 0;
static int workerThread = 0;

static unsigned dimension_x, dimension_y;

static unsigned video_mode;

static Boolean bQuit = FALSE;

#define KBD_RSHIFT_DOWN     0x0001
#define KBD_LSHIFT_DOWN     0x0002
#define KBD_CTRL_DOWN       0x0004
#define KBD_ALT_DOWN        0x0008
#define KBD_SCROLLLOCK_ON   0x0010
#define KBD_NUMLOCK_ON      0x0020
#define KBD_CAPSLOCK_ON     0x0040
#define KBD_INSERT_ON       0x0080
#define KBD_LCTRL_DOWN      0x0100
#define KBD_LALT_DOWN       0x0200
#define KBD_RCTRL_DOWN      0x0400
#define KBD_RALT_DOWN       0x0800
#define KBD_SCROLLLOCK_DOWN 0x1000
#define KBD_NUMLOCK_DOWN    0x2000
#define KBD_CAPSLOCK_DOWN   0x4000
#define KBD_SYSREQ_DOWN     0x8000

#define IS_KBD_PRESS(old,new,flag) (((old&flag)!=flag)&&((new&flag)==flag))
#define IS_KBD_REL(old,new,flag)   (((old&flag)==flag)&&((new&flag)!=flag))

static HMOU hMouse;

enum 
{
        T640X400_25,
        T640X400_50,
        G320X200,
        G640X480,
        G320X240,
        G320X400,
        G640X350,
        G320X96,
        G640X200
};

struct S_MODEINFO
{
    UCHAR    fbtype;
    USHORT   BytesPerScanline;
    USHORT   Width;
    USHORT   Height;
    UCHAR    CharacterWidth;
    UCHAR    CharacterHeight;
    UCHAR    NumberOfBitsPerPixel;
};

static struct S_MODEINFO ModeInfo[] =
{
  {  0x1,  160,  640,  400, 8, 16,  4 },
  {  0x1,  160,  640,  400, 8,  8,  4 },
  {  0x3,  320,  320,  200, 8,  8,  8 },
  {  0xb,  640,  640,  480, 8, 16,  8 },
  {  0xb,  320,  320,  240, 8,  8,  8 },
  {  0xb,  640,  640,  400, 8, 16,  8 },
  {  0xb,  640,  640,  350, 8, 16,  8 },
  {  0xb,  320,  320,   96, 8,  8,  8 },
  {  0xb,  640,  640,  200, 8,  8,  8 },
};

static struct S_MODEINFO *CurrMode;

static PUCHAR vesa_memory;
static int bUseVesa;

static HFILE ScreenHandle = 0;

static void AddScanCode( Bit32u key )
{
   keyevents [ tail ] = key ;
   tail = ( tail + 1 ) % SCANCODE_BUFSIZE ;
}

static Boolean bAllowScreenUpdates = TRUE;

static VOID _Optlink VideoStateThread( PVOID pvoid )
{
   while ( !bQuit )
      {
      USHORT usRestoreScreen;

      if ( VioSavRedrawWait( 0, &usRestoreScreen, 0 ) == 0 )
         {
         bAllowScreenUpdates = ( usRestoreScreen == 1 );
         }
      }

   _endthread();
}

static VOID _Optlink KbdThread( PVOID pvoid )
{
   USHORT usState = 0 ;

   while ( !bQuit )
      {
      KBDKEYINFO kki = {0};

      KbdCharIn( &kki, IO_WAIT, 0 );

      if ( IS_KBD_PRESS(usState,kki.fsState,KBD_RSHIFT_DOWN) )
         AddScanCode(0x36);
      if ( IS_KBD_PRESS(usState,kki.fsState,KBD_LSHIFT_DOWN) )
         AddScanCode(0x2A);
      if ( IS_KBD_PRESS(usState,kki.fsState,KBD_CTRL_DOWN) )
         AddScanCode(0x1D);
      if ( IS_KBD_PRESS(usState,kki.fsState,KBD_ALT_DOWN) )
         AddScanCode(0x38);
      if ( IS_KBD_PRESS(usState,kki.fsState,KBD_SCROLLLOCK_DOWN) )
         AddScanCode(0x46);
      if ( IS_KBD_PRESS(usState,kki.fsState,KBD_NUMLOCK_DOWN) )
         AddScanCode(0x45);
      if ( IS_KBD_PRESS(usState,kki.fsState,KBD_CAPSLOCK_DOWN) )
         AddScanCode(0x3A);
         
      if ( kki.chScan > 0 )
         {
         AddScanCode(kki.chScan);
         AddScanCode(kki.chScan|0x80);
         }

      if ( IS_KBD_REL(usState,kki.fsState,KBD_RSHIFT_DOWN) )
         AddScanCode(0x36|0x80);
      if ( IS_KBD_REL(usState,kki.fsState,KBD_LSHIFT_DOWN) )
         AddScanCode(0x2A|0x80);
      if ( IS_KBD_REL(usState,kki.fsState,KBD_CTRL_DOWN) )
         AddScanCode(0x1D|0x80);
      if ( IS_KBD_REL(usState,kki.fsState,KBD_ALT_DOWN) )
         AddScanCode(0x38|0x80);
      if ( IS_KBD_REL(usState,kki.fsState,KBD_SCROLLLOCK_DOWN) )
         AddScanCode(0x46|0x80);
      if ( IS_KBD_REL(usState,kki.fsState,KBD_NUMLOCK_DOWN) )
         AddScanCode(0x45|0x80);
      if ( IS_KBD_REL(usState,kki.fsState,KBD_CAPSLOCK_DOWN) )
         AddScanCode(0x3A|0x80);

      usState = kki.fsState;
      }

   _endthread();
}

static VOID _Optlink MouThread( PVOID pvoid )
{
   unsigned last_mouse_button = 0;
   USHORT usMask = MOUSE_BN1_DOWN | MOUSE_BN2_DOWN;

   if (MouOpen(NULL, &hMouse) && MouOpen("POINTER$", &hMouse))
      hMouse = 0;

   MouFlushQue( hMouse );

   MouSetEventMask( &usMask, hMouse );

   usMask = 0x100;
   MouSetDevStatus(&usMask, hMouse);

   while ( !bQuit )
      {
      USHORT usWait = MOU_WAIT;
      MOUEVENTINFO mei = {0};
      ULONG ulRC;

      ulRC = MouReadEventQue( &mei, &usWait, hMouse );

      if ( ( mei.fs & MOUSE_BN1_DOWN ) != 0 && (last_mouse_button & 0x01) == 0 )
         {
         last_mouse_button |= 0x01;
         AddScanCode(LMOUSE_PRESSED);
//         gui_printf("mouse button 1 down\n");
         }

      if ( ( mei.fs & MOUSE_BN1_DOWN ) == 0 && (last_mouse_button & 0x01) != 0 )
         {
         last_mouse_button &= ~0x01;
         AddScanCode(LMOUSE_RELEASED);
//         gui_printf("mouse button 1 up\n");
         }

      if ( ( mei.fs & MOUSE_BN2_DOWN ) != 0 && (last_mouse_button & 0x02) == 0 )
         {
         last_mouse_button |= 0x02;
         AddScanCode(RMOUSE_PRESSED);
//         gui_printf("mouse button 2 down\n");
         }

      if ( ( mei.fs & MOUSE_BN2_DOWN ) == 0 && (last_mouse_button & 0x02) != 0 )
         {
         last_mouse_button &= ~0x02;
         AddScanCode(RMOUSE_RELEASED);
//         gui_printf("mouse button 2 up\n");
         }
      }

   if ( hMouse )
      MouClose( hMouse );

   _endthread();
}

void APIENTRY gui_specific_init(bx_gui_c *th, int argc, char **argv, unsigned tilewidth, unsigned tileheight, unsigned headerbar_y)
{
   ULONG action;
   APIRET apirc;
   int i;

   apirc = DosOpen("SCREEN$", &ScreenHandle, &action, 0, 0,
                   OPEN_ACTION_OPEN_IF_EXISTS, OPEN_FLAGS_NOINHERIT |
                   OPEN_SHARE_DENYNONE | OPEN_ACCESS_READWRITE, NULL) ;

   bUseVesa = FALSE;

   // text mode dimensions
   dimension_x = 640;
   dimension_y = 400;
   
   if ( _beginthread (KbdThread, 0, 32768, 0) == -1 )
      {
      exit(1);
      }

   if ( _beginthread (MouThread, 0, 32768, 0) == -1 )
      {
      exit(1);
      }
}

void APIENTRY gui_handle_events(void)
{
   static unsigned ms_prev_x=0;
   static unsigned ms_prev_y=0;

   PTRLOC mousepos;
   Bit32u key;
   unsigned char scancode;
   unsigned ms_curr_x=0;
   unsigned ms_curr_y=0;
   unsigned ms_xdelta;
   unsigned ms_ydelta;
  

   MouGetPtrPos(&mousepos, hMouse);
   ms_curr_y = ( CurrMode ? CurrMode->Width : 80 ) - mousepos.row;
   ms_curr_x = mousepos.col;

   ms_xdelta = ms_curr_x - ms_prev_x;
   ms_ydelta = ms_curr_y - ms_prev_y;
   ms_prev_x = ms_curr_x;
   ms_prev_y = ms_curr_y;

   // Handle mouse moves
   if (ms_xdelta || ms_ydelta)
      {
      gui_mouse_motion(ms_xdelta, ms_ydelta, mouse_button_state);
      }
  
  // Handle keyboard and mouse clicks
   while (head != tail)
      {
      key = keyevents[head];
      head = (head + 1) % SCANCODE_BUFSIZE;

      // Check for mouse buttons first
      if ((key & LMOUSE_PRESSED) || (key & LMOUSE_RELEASED) || (key & RMOUSE_PRESSED) || (key & RMOUSE_RELEASED))
         {
         if ( key & LMOUSE_PRESSED )
            {
            mouse_button_state |= 0x01;
            }
         if ( key & LMOUSE_RELEASED )
            {
            mouse_button_state &= ~0x01;
            }
         if ( key & RMOUSE_PRESSED )
            {
            mouse_button_state |= 0x02;
            }
         if ( key & RMOUSE_RELEASED )
            {
            mouse_button_state &= ~0x02;
            }
         gui_mouse_motion( 1, 1, mouse_button_state);
         gui_mouse_motion( 0, 0, key);
         gui_mouse_motion( -1, -1, mouse_button_state);
         }
      else
         {
         scancode = key;
         gui_put_scancode(&scancode, 1);
         }
      }
}

void APIENTRY gui_text_update(Bit8u *old_text, Bit8u *new_text, unsigned long cursor_x, unsigned long cursor_y, unsigned nrows)
{
   int i;
   unsigned nchars;

   while ( !bAllowScreenUpdates )
      DosSleep( 10 );

   nchars = 80 * nrows;

   for (i=0; i<nchars*2; i+=2)
      {
      if ((old_text[i] != new_text[i]) || (old_text[i+1] != new_text[i+1]))
         {
         VioWrtCellStr( (char *)&new_text[i], 2, (i/2) / 80, (i/2) % 80, 0 );
         }
      }

   VioSetCurPos( cursor_y, cursor_x, 0 );
}

Boolean APIENTRY gui_palette_change(unsigned index, unsigned red, unsigned green, unsigned blue)
{
   while ( !bAllowScreenUpdates )
      DosSleep( 10 );

   _outb1( 0x3c8, index );
   _outb1( 0x3c9, red >> 2 );
   _outb1( 0x3c9, green >> 2 );
   _outb1( 0x3c9, blue >> 2 );

   return 0;
}

void APIENTRY gui_graphics_tile_update(Bit8u *tile, unsigned x, unsigned y)
{
   int i, j, offset;
   PUCHAR video;

   while ( !bAllowScreenUpdates )
      DosSleep( 10 );

   video = (bUseVesa) ? vesa_memory : pbx_mem->video;

   if ( video_mode != G320X400 )
      {
      offset = y * CurrMode->BytesPerScanline + x ;

      for ( i = 0 ; i < 16 && y+i < CurrMode->Height ; i++ )
         {
         memcpy( &video[offset], tile, 16 );
         offset += CurrMode->BytesPerScanline;
         tile += 16;
         }
      }
   else
      {
      offset = y * CurrMode->BytesPerScanline + x * 2 ;

      for ( i = 0 ; i < 16 && y+i < CurrMode->Height ; i++ )
         {
         for ( j = 0 ; j < 16 ; j++ )
            {
            video[offset + j*2] = (UCHAR) tile[j];
            video[offset + j*2+1] = (UCHAR) tile[j];
            }
         offset += CurrMode->BytesPerScanline;
         tile += 16;
         }
      }
}

void APIENTRY gui_dimension_update(unsigned x, unsigned y)
{
   while ( !bAllowScreenUpdates )
      DosSleep( 10 );

   if ( dimension_x != x || dimension_y != y )
      {
      VIOMODEINFO VioModeInfo;
      int i;
 
      switch ( x )
      {
         case 320:
            switch ( y )
            {
               case  96: video_mode = G320X96;  break;
               case 208: video_mode = G320X200; break;
               case 240: video_mode = G320X240; break;
               case 400: video_mode = G320X400; break;
            }
            break;

         case 640:
            switch ( y )
            {
               case 208: video_mode = G640X200; break;
               case 352: video_mode = G640X350; break;
               case 400: video_mode = T640X400_25; break;
               case 480: video_mode = G640X480; break;
               case 800: video_mode = T640X400_50; break;
            }
            break;
      }

      CurrMode = &ModeInfo[video_mode];

      VioModeInfo.cb     = 12;
      VioModeInfo.fbType = CurrMode->fbtype;
      VioModeInfo.color  = CurrMode->NumberOfBitsPerPixel;
      VioModeInfo.col    = CurrMode->Width  / CurrMode->CharacterWidth;
      VioModeInfo.row    = CurrMode->Height / CurrMode->CharacterHeight;
      VioModeInfo.hres   = CurrMode->Width;
      VioModeInfo.vres   = CurrMode->Height;
  
      VioSetMode(&VioModeInfo, 0);

      VioModeInfo.cb = sizeof(VIOMODEINFO);
      VioGetMode(&VioModeInfo, 0);

      if ( VioModeInfo.buf_addr > 0x100000 )
         {
         if ( !vesa_memory )
            {
            vesa_memory = (UCHAR *)create_pmap( VioModeInfo.buf_addr, VioModeInfo.buf_length );
            }
         bUseVesa = TRUE;
         }
      else
         {
         bUseVesa = FALSE;
         }

      dimension_x = x;
      dimension_y = y;
      }
}

void APIENTRY gui_exit(void)
{
   bQuit = TRUE;

   gui_dimension_update(640, 400);

   if (ScreenHandle != 0)
      DosClose(ScreenHandle);
}
