// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
//
//    US Mail:  Kevin Lawton
//              528 Lexington St.
//              Waltham, MA 02154
//
//    EMail:    bochs@world.std.com

//  Much of this file was written by:
//  Craig S Ballantyne

#define INCL_DOS
#define INCL_DEV
#define INCL_GPI
#define INCL_WIN
#include "bochs.h"

#include "../../font/vga.bitmap.h"

#include <process.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#include "..\..\os2\os2cp.h"

#define WM_SET_NEXT_EVENT  WM_USER+1
#define WM_GET_NEXT_EVENT  WM_USER+2
#define WM_CREATE_BITMAP   WM_USER+3
#define WM_DISPLAY_BITMAP  WM_USER+4
#define WM_REPAINT_BITMAP  WM_USER+5
#define WM_CLEAR_SCREEN    WM_USER+6
#define WM_RESIZE_SCREEN   WM_USER+7
#define WM_PALETTE_CHANGE  WM_USER+8
#define WM_TILE_CHANGE     WM_USER+9

#define TILESIZE_X   16
#define TILESIZE_Y   16
#define VGA_X        640
#define VGA_Y        480

#ifdef __EMX__
#define min(a, b)       (((a) < (b))? (a) : (b))
#define max(a, b)       (((a) > (b))? (a) : (b))
#endif

// Keyboard/mouse stuff
#define SCANCODE_BUFSIZE    100
#define LMOUSE_PRESSED      0x20000000
#define RMOUSE_PRESSED      0x40000000
#define LMOUSE_RELEASED     0x02000000
#define RMOUSE_RELEASED     0x04000000

static Bit32u keyevents[SCANCODE_BUFSIZE];
static unsigned screen_updated=0;
static unsigned head=0, tail=0;
static unsigned mouse_button_state = 0;
static unsigned ms_curr_x=0, ms_curr_y=0, ms_prev_x=0, ms_prev_y=0;
static int workerThread = 0;

// Graphics screen stuff
static BITMAPINFO2 *pbm256;
static char        *pVGA;
static HBITMAP  hbmVGA[256];
static HBITMAP  hbmInverseVGA[256];
static RECTL    rclScreen;

static unsigned bx_bitmap_entries = 0;
static unsigned bx_bitmap_left_xorigin = 0;   // pixels from left
static unsigned bx_bitmap_right_xorigin = 0;  // pixels from right

// Misc stuff
static unsigned dimension_x, dimension_y;
static unsigned window_x, window_y;
static unsigned prev_block_cursor_x = 0;
static unsigned prev_block_cursor_y = 0;
static unsigned y_edge=0, x_edge=0, y_caption=0;

static char *pszAppName = "Bochs/2";
static char *pszWindowName = "Bochs/2 VGA Monitor";

typedef struct
{
   char    *kill;  // reason for gui_terminate(int)
   BOOL     UIinited;
   HWND     hwnd;
   HWND     frame;
   HWND     dialog;
} sharedThreadInfo;

static sharedThreadInfo stInfo;

// *********************************************************************
// BOCHS General Functions
// *********************************************************************

extern void gui_put_scancode ( PUCHAR puc, INT i );
extern void gui_mouse_motion (INT i1, INT i2, UINT ui);
extern void gui_terminate(char *reason);
extern HWND gui_control_panel(HWND hwnd);

// *********************************************************************
// Display Window Functions
// *********************************************************************

static MRESULT APIENTRY wpBochsDisplay(HWND hwnd, ULONG iMsg, MPARAM mp1, MPARAM mp2)
{
   static HDC      MemoryDC;
   static HPS      MemoryPS;
   static HBITMAP  MemoryBM;
   static HBITMAP  TileBM;

   HPS               hps;
   POINTL            aptl[4];
   RECTL             rcl;
   SIZEL             sizel;
   BITMAPINFOHEADER2 bmh;
   HRGN              hrgn;
   Bit32u            key;
   Bit32u            press_release;
   BITMAPINFO2      *pbm;
   HBITMAP           hBitmap;
   int               x, y, cx, cy;
   unsigned char    *data;
   int               index;
   RGB2              rgb;
   int               i, j;
   Bit8u            *tile;
   unsigned          pos_x;
   unsigned          pos_y;
   RECTL             rclDialog = {0};


   switch (iMsg)
      {
      case WM_CREATE:
         dimension_x = 0;
         dimension_y = 0;
  
         x_edge      = WinQuerySysValue(HWND_DESKTOP,SV_CXDLGFRAME);
         y_edge      = WinQuerySysValue(HWND_DESKTOP,SV_CYDLGFRAME);
         y_caption   = WinQuerySysValue(HWND_DESKTOP,SV_CYTITLEBAR);

         window_x = 2*x_edge;
         window_y = y_caption + 2*y_edge;

         pVGA = (char *)calloc( VGA_X, VGA_Y );

         hps = WinGetPS( hwnd );

         MemoryDC = DevOpenDC( NULL, OD_MEMORY, "*", 0, 0, hps );

         sizel.cx = VGA_X;
         sizel.cy = VGA_Y;
         MemoryPS = GpiCreatePS( NULL, MemoryDC, &sizel, PU_PELS | GPIF_DEFAULT | GPIA_ASSOC | GPIT_NORMAL );

         bmh.cbFix     = sizeof(BITMAPINFOHEADER2);
         bmh.cx        = VGA_X;
         bmh.cy        = VGA_Y;
         bmh.cPlanes   = 1;
         bmh.cBitCount = 8;
         TileBM = GpiCreateBitmap(MemoryPS,&bmh,0,0,0);

         bmh.cbFix     = sizeof(BITMAPINFOHEADER2);
         bmh.cx        = VGA_X;
         bmh.cy        = VGA_Y;
         bmh.cPlanes   = 1;
         bmh.cBitCount = 8;
         MemoryBM = GpiCreateBitmap(MemoryPS,&bmh,0,0,0);
         GpiSetBitmap( MemoryPS, MemoryBM );

         WinReleasePS( hps );

         WinSendMsg( hwnd, WM_CLEAR_SCREEN, MPFROM2SHORT(0,0), MPFROM2SHORT(VGA_X,VGA_Y) );

         stInfo.UIinited = TRUE;

         stInfo.dialog = gui_control_panel( hwnd );
         break;

      case WM_CLOSE:
         WinDismissDlg( stInfo.dialog, 0 );
         WinPostMsg( hwnd, WM_QUIT, 0, 0 );
         break;

      case WM_DESTROY:
         GpiDeleteBitmap( TileBM );
         GpiDeleteBitmap( MemoryBM );
         GpiAssociate( MemoryPS, NULL );
         GpiDestroyPS( MemoryPS );
         DevCloseDC( MemoryDC );
         break;

      case WM_PAINT:
         hps = WinBeginPaint (hwnd, NULL, &rcl);
         aptl[0].x = rcl.xLeft;
         aptl[0].y = rcl.yBottom;
         aptl[1].x = rcl.xRight;
         aptl[1].y = rcl.yTop;
         aptl[2].x = rcl.xLeft;
         aptl[2].y = rcl.yBottom;
         aptl[3].x = rcl.xRight;
         aptl[3].y = rcl.yTop;
         GpiBitBlt(hps, MemoryPS, 3, aptl, ROP_SRCCOPY, BBO_IGNORE);
         WinEndPaint (hwnd);
         return (MRESULT) 1;

      case WM_MOUSEMOVE:
         ms_curr_x = SHORT1FROMMP(mp1);
         ms_curr_y = SHORT2FROMMP(mp1);
         break;

      case WM_BUTTON1DOWN:
         mouse_button_state |= 0x01;
         WinSendMsg( hwnd, WM_SET_NEXT_EVENT, MPFROMLONG(mouse_button_state), MPFROMLONG(LMOUSE_PRESSED) );
         break;
    
      case WM_BUTTON1UP:
         mouse_button_state &= ~0x01;
         WinSendMsg( hwnd, WM_SET_NEXT_EVENT, MPFROMLONG(mouse_button_state), MPFROMLONG(LMOUSE_RELEASED) );
         break;

      case WM_BUTTON2DOWN:
         mouse_button_state |= 0x02;
         WinSendMsg( hwnd, WM_SET_NEXT_EVENT, MPFROMLONG(mouse_button_state), MPFROMLONG(RMOUSE_PRESSED) );
         break;
    
      case WM_BUTTON2UP:
         mouse_button_state &= ~0x02;
         WinSendMsg( hwnd, WM_SET_NEXT_EVENT, MPFROMLONG(mouse_button_state), MPFROMLONG(RMOUSE_RELEASED) );
         break;
    
      case WM_CHAR:
         if ( SHORT1FROMMP(mp1) & KC_SCANCODE )
            {
            unsigned char scancode = CHAR4FROMMP(mp1);

            switch( scancode )
               {
               case 0x5a:  scancode = 0x1c;  break;
               case 0x5c:  scancode = 0x35;  break;
               case 0x60:  scancode = 0x47;  break;
               case 0x61:  scancode = 0x48;  break;
               case 0x62:  scancode = 0x49;  break;
               case 0x63:  scancode = 0x4b;  break;
               case 0x64:  scancode = 0x4d;  break;
               case 0x65:  scancode = 0x4f;  break;
               case 0x66:  scancode = 0x50;  break;
               case 0x67:  scancode = 0x51;  break;
               case 0x68:  scancode = 0x52;  break;
               case 0x69:  scancode = 0x53;  break;
               }

            if ( SHORT1FROMMP(mp1) & KC_KEYUP )
               {
               WinSendMsg( hwnd, WM_SET_NEXT_EVENT, MPFROMLONG(scancode), MPFROMLONG(BX_KEY_RELEASED) );
               }
            else
               {
               WinSendMsg( hwnd, WM_SET_NEXT_EVENT, MPFROMLONG(scancode), MPFROMLONG(BX_KEY_PRESSED) );
               }
            }
         return (MRESULT) 1;

      case WM_SET_NEXT_EVENT:
         key = LONGFROMMP(mp1);
         press_release = LONGFROMMP(mp2);

         if (((tail+1) % SCANCODE_BUFSIZE) == head)
            {
            fprintf(stderr, "# enq_scancode: buffer full\n");
            }
         else
            {
            keyevents[tail] = key | press_release;
            tail = (tail + 1) % SCANCODE_BUFSIZE;
            }
         break;

      case WM_GET_NEXT_EVENT:
         if ( head == tail )
            {
            fprintf(stderr, "# deq_scancode: buffer empty\n");
            return(0);
            }
         else
            {
            key = keyevents[head];
            head = (head + 1) % SCANCODE_BUFSIZE;
            return (MRESULT)key;
            }
         break;

      case WM_CREATE_BITMAP:
         cx    = SHORT1FROMMP( mp1 );
         cy    = SHORT2FROMMP( mp1 );
         data  = (unsigned char *) mp2;

         pbm   = (BITMAPINFO2 *) calloc( 1, sizeof(BITMAPINFO2) + sizeof(RGB2) );

         pbm->cbFix     = sizeof(BITMAPINFOHEADER2);
         pbm->cx        = cx;
         pbm->cy        = cy;
         pbm->cPlanes   = 1;
         pbm->cBitCount = 1;

         pbm->argbColor[0].bBlue     = 0;
         pbm->argbColor[0].bGreen    = 0;
         pbm->argbColor[0].bRed      = 0;
         pbm->argbColor[0].fcOptions = 0;

         pbm->argbColor[1].bBlue     = 0xff;
         pbm->argbColor[1].bGreen    = 0xff;
         pbm->argbColor[1].bRed      = 0xff;
         pbm->argbColor[1].fcOptions = 0;

         hBitmap = GpiCreateBitmap(MemoryPS,(BITMAPINFOHEADER2*)pbm,0,0,0);

         GpiSetBitmap( MemoryPS, hBitmap );

         for ( i = 0 ; i < cy ; i++ )
            {
            char acData[32] = {0};
            memcpy( acData, &data[(cy-i-1)*cx/8], cx/8 );
            GpiSetBitmapBits( MemoryPS, i, 1, acData, pbm );
            }

         free(pbm);

         GpiSetBitmap( MemoryPS, MemoryBM );

         return (MRESULT) hBitmap;

      case WM_DISPLAY_BITMAP:
         x        = SHORT1FROMMP( mp1 );
         y        = SHORT2FROMMP( mp1 );
         hBitmap  = (HBITMAP) mp2;
         GpiQueryBitmapDimension( hBitmap, &sizel );
         cx       = sizel.cx;
         cy       = sizel.cy;

         aptl[0].x = x;
         aptl[0].y = y;
         aptl[1].x = cx;
         aptl[1].y = cy;
         aptl[2].x = 0;
         aptl[2].y = 0;
         aptl[3].x = cx;
         aptl[3].y = cy;
         WinDrawBitmap( MemoryPS, hBitmap, NULL, aptl, 0, 0, DBM_NORMAL | DBM_IMAGEATTRS );
         break;

      case WM_REPAINT_BITMAP:
         pbm256->cx = dimension_x;
         pbm256->cy = dimension_y;

         GpiSetBitmap( MemoryPS, TileBM );

         for ( i = 0 ; i < dimension_y ; i++ )
            GpiSetBitmapBits( MemoryPS, i, 1, &pVGA[dimension_x*(dimension_y-i-1)], pbm256 );

         GpiSetBitmap( MemoryPS, MemoryBM );

         aptl[0].x = 0;
         aptl[0].y = 0;
         aptl[1].x = dimension_x;
         aptl[1].y = dimension_y;
         aptl[2].x = 0;
         aptl[2].y = 0;
         aptl[3].x = dimension_x;
         aptl[3].y = dimension_y;
         WinDrawBitmap( MemoryPS, TileBM, NULL, aptl, 0, 0, DBM_NORMAL | DBM_IMAGEATTRS );

         WinInvalidateRect( hwnd, &rclScreen, FALSE );
         break;

      case WM_CLEAR_SCREEN:
         rcl.xLeft   = SHORT1FROMMP(mp1);
         rcl.xRight  = SHORT1FROMMP(mp2);
         rcl.yBottom = SHORT2FROMMP(mp1);
         rcl.yTop    = SHORT2FROMMP(mp2);

         hrgn = GpiCreateRegion( MemoryPS, 1, &rcl );
         GpiSetColor( MemoryPS, CLR_PALEGRAY );
         GpiPaintRegion( MemoryPS, hrgn );
         GpiDestroyRegion( MemoryPS, hrgn );
         GpiSetColor( MemoryPS, CLR_BLACK );

         WinInvalidateRect( hwnd, &rcl, FALSE );
         break;

      case WM_RESIZE_SCREEN:
         memset( pVGA, 0, 640 * 480 );
         
         dimension_x = SHORT1FROMMP(mp1);
         dimension_y = SHORT2FROMMP(mp1);

         window_x = dimension_x + 2*x_edge;
         window_y = dimension_y + y_caption + 2*y_edge;

         if ( stInfo.dialog )
            WinQueryWindowRect ( stInfo.dialog, &rclDialog );

         pos_x  = (WinQuerySysValue(HWND_DESKTOP,SV_CXSCREEN)-window_x)/2;
         pos_y  = (WinQuerySysValue(HWND_DESKTOP,SV_CYSCREEN)-window_y)/2 + (rclDialog.yTop - rclDialog.yBottom)/2;
         WinSetWindowPos(stInfo.frame,HWND_TOP,pos_x,pos_y,window_x,window_y,SWP_SIZE | SWP_MOVE );

         pos_y  = (WinQuerySysValue(HWND_DESKTOP,SV_CYSCREEN)-window_y)/2 - (rclDialog.yTop - rclDialog.yBottom)/2;
         WinSetWindowPos(stInfo.dialog,HWND_TOP,pos_x,pos_y,0,0,SWP_MOVE );

         break;

      case WM_PALETTE_CHANGE:
         index = LONGFROMMP(mp1);
         memcpy( &rgb, &mp2, sizeof(long) );

         if ( !pbm256 )
            {
            pbm256 = (BITMAPINFO2 *) calloc( 1, sizeof(BITMAPINFO2) + 256 * sizeof(RGB2) );

            pbm256->cbFix     = sizeof(BITMAPINFOHEADER2);
            pbm256->cPlanes   = 1;
            pbm256->cBitCount = 8;
            }

         pbm256->argbColor[index].bBlue     = rgb.bBlue;
         pbm256->argbColor[index].bGreen    = rgb.bGreen;
         pbm256->argbColor[index].bRed      = rgb.bRed;
         pbm256->argbColor[index].fcOptions = 0;
         break;

      case WM_TILE_CHANGE:
         x     = SHORT1FROMMP( mp1 );
         y     = SHORT2FROMMP( mp1 );
         tile  = (Bit8u *) mp2;

         for ( i = 0 ; i < TILESIZE_Y ; i++ )
            {
            j = ( y + i ) * dimension_x + x;
            memcpy( &pVGA[j], &tile[i*TILESIZE_X], TILESIZE_X );
            }
         break;
   }

   return WinDefWindowProc (hwnd, iMsg, mp1, mp2);
}

// This thread controls the GUI window.
static void _Optlink UIThread(void *pvoid)
{
   HAB   hab;
   HMQ   hmq;
   QMSG  qmsg;
   ULONG flCreate;

   hab = WinInitialize(0);
   hmq = WinCreateMsgQueue(hab,0);

   WinRegisterClass(hab,pszAppName,wpBochsDisplay,0,sizeof(LONG));

   flCreate = FCF_TASKLIST | FCF_TITLEBAR | FCF_MINBUTTON | FCF_DLGBORDER ;

   stInfo.frame = WinCreateStdWindow(HWND_DESKTOP,
      WS_VISIBLE,
      &flCreate,
      pszAppName,
      pszWindowName,
      WS_VISIBLE,
      0,
      666,
      &stInfo.hwnd );

   if (!stInfo.frame) gui_terminate("window creation error!\n");

   WinSetFocus( HWND_DESKTOP, stInfo.hwnd );

   while (WinGetMsg (hab, &qmsg, 0, 0, 0))
      WinDispatchMsg (hab, &qmsg);

   WinDestroyWindow( stInfo.frame );
   WinDestroyMsgQueue( hmq );
   WinTerminate( hab );

   stInfo.kill = "Window closed, exiting!\n";

   _endthread();
}

static unsigned char reverse_bitorder(unsigned char b)
{
   unsigned char ret=0;
	
   for (unsigned i=0; i<8; i++)
      {
      ret |= (b & 0x01) << (7-i);
      b >>= 1;
      }
	
   return(ret);
}

static void DisplayBitmap(unsigned x, unsigned y, unsigned cx, unsigned cy, unsigned char *data, HBITMAP *phBitmap)
{
   if ( !*phBitmap )
      *phBitmap = (HBITMAP) WinSendMsg( stInfo.hwnd, WM_CREATE_BITMAP, MPFROM2SHORT(cx,cy), MPFROMP(data) );
   WinSendMsg( stInfo.hwnd, WM_DISPLAY_BITMAP, MPFROM2SHORT(x,y), MPFROMLONG(*phBitmap) );
}

static void DisplayFont(unsigned c, unsigned x, unsigned y, BOOL bNormal)
{
   unsigned char data[16];

   for (unsigned i=0; i<16; i++)
      data[i] = reverse_bitorder(bNormal ? ~bx_vgafont[c].data[i] : bx_vgafont[c].data[i]);

   y = 24 * 16 - y;

   DisplayBitmap(x,y,8,16,data,bNormal?&hbmVGA[c]:&hbmInverseVGA[c]);

   rclScreen.xLeft   = min(rclScreen.xLeft,x);
   rclScreen.xRight  = max(rclScreen.xRight,x+8);
   rclScreen.yBottom = min(rclScreen.yBottom,y);
   rclScreen.yTop    = max(rclScreen.yTop,y+16);
}

void APIENTRY gui_specific_init(bx_gui_c *th, int argc, char **argv, unsigned tilewidth, unsigned tileheight, unsigned headerbar_y)
{
   HAB   hab;
   HMQ   hmq;

   hab = WinInitialize(0);
   hmq = WinCreateMsgQueue(hab,0);

   stInfo.kill = 0;
   stInfo.UIinited = FALSE;

   workerThread = _beginthread (UIThread, 0, 32768, 0);

   if (workerThread == -1)
      gui_terminate("thread creation error!\n");

   // Wait for a window before continuing
   while (!stInfo.UIinited) DosSleep(10);
}

void APIENTRY gui_handle_events(void)
{
   Bit32u key;
   unsigned char scancode;
   unsigned ms_xdelta = ms_curr_x - ms_prev_x;
   unsigned ms_ydelta = ms_curr_y - ms_prev_y;
  
   // printf("# Hey!!!\n");

   if (stInfo.kill) gui_terminate(stInfo.kill);
  
   if ( screen_updated )
      {
      WinSendMsg( stInfo.hwnd, WM_REPAINT_BITMAP, 0, 0 );
      screen_updated = 0;
      }

   ms_prev_x = ms_curr_x;
   ms_prev_y = ms_curr_y;

   // Handle mouse moves
   if (ms_xdelta || ms_ydelta)
      {
      gui_mouse_motion(ms_xdelta, ms_ydelta, mouse_button_state);
      }
  
  // Handle keyboard and mouse clicks
   while (head != tail)
      {
      // Bypass bx_devices.keyboard->gen_scancode so we may enter
      // a scancode directly bx_devices.keyboard->gen_scancode();
      key = (Bit32u) WinSendMsg( stInfo.hwnd, WM_GET_NEXT_EVENT, 0, 0 );

      // Check for mouse buttons first
      if ((key & LMOUSE_PRESSED) || (key & LMOUSE_RELEASED) || (key & RMOUSE_PRESSED) || (key & RMOUSE_RELEASED))
         {
         // printf("# click!\n");
         gui_mouse_motion( 1, 1, mouse_button_state);
         gui_mouse_motion( 0, 0, key);
         gui_mouse_motion( -1, -1, mouse_button_state);
         }
      else
         {
         // Its a key
         // printf("# key = %d, scancode = %d\n",key,scancode);
         if (key & BX_KEY_RELEASED)
            {
            scancode = key - BX_KEY_RELEASED;
            scancode |= 0x80;
            }
         else
            {
            scancode = key;
            }

         gui_put_scancode(&scancode, 1);
         }
      }
}

void APIENTRY gui_clear_screen(void)
{
   WinSendMsg( stInfo.hwnd, WM_CLEAR_SCREEN, MPFROM2SHORT(0,0), MPFROM2SHORT(dimension_x,dimension_y) );
}

void APIENTRY gui_text_update(Bit8u *old_text, Bit8u *new_text, unsigned long cursor_x, unsigned long cursor_y, unsigned nrows)
{
   unsigned char achar;
   unsigned i, x, y;
   unsigned nchars;

   if (!stInfo.UIinited) return;
	
   rclScreen.xLeft   = dimension_x;
   rclScreen.xRight  = 0;
   rclScreen.yBottom = dimension_y;
   rclScreen.yTop    = 0;
  
   nchars = 80 * nrows;

   if ( (prev_block_cursor_y*80 + prev_block_cursor_x) < nchars)
      {
      achar = new_text[(prev_block_cursor_y*80 + prev_block_cursor_x)*2];
      DisplayFont(achar, prev_block_cursor_x*8, prev_block_cursor_y*16, TRUE);
      }
  
   for (i=0; i<nchars*2; i+=2)
      {
      if ((old_text[i] != new_text[i]) || (old_text[i+1] != new_text[i+1]))
         {
         achar = new_text[i];

         x = (i/2) % 80;
         y = (i/2) / 80;

         DisplayFont(achar, x*8, y*16, TRUE);
         }
      }
  
   prev_block_cursor_x = cursor_x;
   prev_block_cursor_y = cursor_y;
  
   // now draw character at new block cursor location in reverse
   if ((cursor_y*80 + cursor_x) < nchars )
      {
      achar = new_text[(cursor_y*80 + cursor_x)*2];
      DisplayFont(achar, cursor_x*8, cursor_y*16, FALSE);
      }

   WinInvalidateRect( stInfo.hwnd, &rclScreen, FALSE );
}

Boolean APIENTRY gui_palette_change(unsigned index, unsigned red, unsigned green, unsigned blue)
{
   RGB2     rgb;
   MPARAM   mp2;

   rgb.bBlue     = blue;
   rgb.bGreen    = green;
   rgb.bRed      = red;
   rgb.fcOptions = 0;

   memcpy( &mp2, &rgb, sizeof(long) );

   WinSendMsg( stInfo.hwnd, WM_PALETTE_CHANGE, MPFROMSHORT(index), mp2 );

   return 0;
}

void APIENTRY gui_graphics_tile_update(Bit8u *tile, unsigned x, unsigned y)
{
   WinSendMsg( stInfo.hwnd, WM_TILE_CHANGE, MPFROM2SHORT(x,y), MPFROMP(tile) );

   screen_updated = 1;

   rclScreen.xLeft   = min(rclScreen.xLeft,x);
   rclScreen.xRight  = max(rclScreen.xRight,x+TILESIZE_X);
   rclScreen.yBottom = min(rclScreen.yBottom,dimension_y-y-TILESIZE_Y);
   rclScreen.yTop    = max(rclScreen.yTop,dimension_y-y);
}

void APIENTRY gui_dimension_update(unsigned x, unsigned y)
{
   BOOL  bRedrawHeader;

   bRedrawHeader = ( dimension_x != x || dimension_y != y );

   if ( bRedrawHeader )
      {
      WinSendMsg( stInfo.hwnd, WM_RESIZE_SCREEN, MPFROM2SHORT(x,y), 0 );

      WinSendMsg( stInfo.hwnd, WM_CLEAR_SCREEN, MPFROM2SHORT(0,0), MPFROM2SHORT(window_x,window_y) );
      
      rclScreen.xLeft   = 0;
      rclScreen.xRight  = dimension_x;
      rclScreen.yBottom = 0;
      rclScreen.yTop    = dimension_y;
      }
   else
      {
      rclScreen.xLeft   = dimension_x;
      rclScreen.xRight  = 0;
      rclScreen.yBottom = dimension_y;
      rclScreen.yTop    = 0;
      }
}

void APIENTRY gui_exit(void)
{
   printf("# In bx_gui_c::exit(void)!\n");
  
   // kill thread first...
   WinPostMsg(stInfo.hwnd, WM_CLOSE, 0, 0);

   // Wait until it dies
   while ((stInfo.kill == 0)) DosSleep(10);

   gui_terminate(NULL);
}

