//  Copyright (C) 2000  MandrakeSoft S.A.
//
//    MandrakeSoft S.A.
//    43, rue d'Aboukir
//    75002 Paris - France
//    http://www.linux-mandrake.com/
//    http://www.mandrakesoft.com/
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA




#include "bochs.h"



#if BX_USE_PAR_SMF
bx_parallel_c bx_parallel;
#define this (&bx_parallel)
#endif


/* 
 * bit defines for 8255 status port
 * base + 1
 * accessed with LP_S(minor), which gets the byte...
 */
#define LP_PBUSY	0x80  /* inverted input, active high */
#define LP_PACK		0x40  /* unchanged input, active low */
#define LP_POUTPA	0x20  /* unchanged input, active high */
#define LP_PSELECD	0x10  /* unchanged input, active high */
#define LP_PERRORP	0x08  /* unchanged input, active low */

/* 
 * defines for 8255 control port
 * base + 2 
 * accessed with LP_C(minor)
 */
#define LP_PINTEN	0x10  /* high to read data in or-ed with data out */
#define LP_PSELECP	0x08  /* inverted output, active low */
#define LP_PINITP	0x04  /* unchanged output, active low */
#define LP_PAUTOLF	0x02  /* inverted output, active low */
#define LP_PSTROBE	0x01  /* short high output on raising edge */

bx_parallel_c::bx_parallel_c(void)
{
  // nothing for now
}

bx_parallel_c::~bx_parallel_c(void)
{
  set_status( 0, FALSE );
  set_status( 1, FALSE );
  set_status( 2, FALSE );
}

  void
bx_parallel_c::init(bx_devices_c *d)
{
  BX_PAR_THIS devices = d;
  Bit32u ret;
  Bit16u portaddr;

  /* PARALLEL PORT 1-3 */

  portaddr = BX_LPT1_IO;
  BX_MEM.write_physical( 0x408, sizeof(portaddr), &portaddr );

  portaddr = BX_LPT2_IO;
  BX_MEM.write_physical( 0x40A, sizeof(portaddr), &portaddr );

  portaddr = BX_LPT3_IO;
  BX_MEM.write_physical( 0x40C, sizeof(portaddr), &portaddr );

  BX_PAR_THIS devices->register_irq(7, "Parallel Port 1");

  BX_PAR_THIS devices->register_io_write_handler(this,
     (bx_write_handler_t) write_handler,
     BX_LPT1_IO, "Parallel Port 1");
  BX_PAR_THIS devices->register_io_read_handler(this,
     (bx_read_handler_t) read_handler,
     BX_LPT1_IO, "Parallel Port 1");

  BX_PAR_THIS devices->register_io_read_handler(this,
     (bx_read_handler_t) read_handler,
     BX_LPT1_IO+1, "Parallel Port 1");

  BX_PAR_THIS devices->register_io_write_handler(this,
     (bx_write_handler_t) write_handler,
     BX_LPT1_IO+2, "Parallel Port 1");
  BX_PAR_THIS devices->register_io_read_handler(this,
     (bx_read_handler_t) read_handler,
     BX_LPT1_IO+2, "Parallel Port 1");

  BX_PAR_THIS devices->register_io_write_handler(this,
     (bx_write_handler_t) write_handler,
     BX_LPT2_IO, "Parallel Port 2");
  BX_PAR_THIS devices->register_io_read_handler(this,
     (bx_read_handler_t) read_handler,
     BX_LPT2_IO, "Parallel Port 2");

  BX_PAR_THIS devices->register_io_read_handler(this,
     (bx_read_handler_t) read_handler,
     BX_LPT2_IO+1, "Parallel Port 2");

  BX_PAR_THIS devices->register_io_write_handler(this,
     (bx_write_handler_t) write_handler,
     BX_LPT2_IO+2, "Parallel Port 2");
  BX_PAR_THIS devices->register_io_read_handler(this,
     (bx_read_handler_t) read_handler,
     BX_LPT2_IO+2, "Parallel Port 2");

  BX_PAR_THIS devices->register_io_write_handler(this,
     (bx_write_handler_t) write_handler,
     BX_LPT3_IO, "Parallel Port 3");
  BX_PAR_THIS devices->register_io_read_handler(this,
     (bx_read_handler_t) read_handler,
     BX_LPT3_IO, "Parallel Port 3");

  BX_PAR_THIS devices->register_io_read_handler(this,
     (bx_read_handler_t) read_handler,
     BX_LPT3_IO+1, "Parallel Port 3");

  BX_PAR_THIS devices->register_io_write_handler(this,
     (bx_write_handler_t) write_handler,
     BX_LPT3_IO+2, "Parallel Port 3");
  BX_PAR_THIS devices->register_io_read_handler(this,
     (bx_read_handler_t) read_handler,
     BX_LPT3_IO+2, "Parallel Port 3");

  set_status( 0, TRUE );
  set_status( 1, TRUE );
  set_status( 2, TRUE );
}

  // static IO port read callback handler
  // redirects to non-static class handler to avoid virtual functions

  Bit32u
bx_parallel_c::read_handler(void *this_ptr, Bit32u address, unsigned io_len)
{
#if !BX_USE_PAR_SMF
  bx_parallel_c *class_ptr = (bx_parallel_c *) this_ptr;

  return( class_ptr->read(address, io_len) );
}


  Bit32u
bx_parallel_c::read(Bit32u address, unsigned io_len)
{
#else
  UNUSED(this_ptr);
#endif  // !BX_USE_PAR_SMF

  /* PARALLEL PORT 1 */
  if (io_len > 1)
    bx_panic("parallel: io read from address %08x, len=%u\n",
             (unsigned) address, (unsigned) io_len);

  if ( address == BX_LPT1_IO )
      {
      return BX_PAR_THIS s[0].cValue;
      }

  if ( address == BX_LPT2_IO )
      {
      return BX_PAR_THIS s[1].cValue;
      }

  if ( address == BX_LPT3_IO )
      {
      return BX_PAR_THIS s[2].cValue;
      }

  if ( address == BX_LPT1_IO+1 || address == BX_LPT2_IO+1 || address == BX_LPT3_IO+1 )
      {
      return LP_PBUSY|LP_PSELECD|LP_PERRORP;
      }

  if ( address == BX_LPT1_IO+2 || address == BX_LPT2_IO+2 || address == BX_LPT3_IO+2 )
      {
      return LP_PSELECP;
      }
      
  return(0);
}


  // static IO port write callback handler
  // redirects to non-static class handler to avoid virtual functions

  void
bx_parallel_c::write_handler(void *this_ptr, Bit32u address, Bit32u value, unsigned io_len)
{
#if !BX_USE_PAR_SMF
  bx_parallel_c *class_ptr = (bx_parallel_c *) this_ptr;

  class_ptr->write(address, value, io_len);
}

  void
bx_parallel_c::write(Bit32u address, Bit32u value, unsigned io_len)
{
#else
  UNUSED(this_ptr);
#endif  // !BX_USE_PAR_SMF
  UNUSED(value);
  UNUSED(io_len);

  /* PARALLEL PORT 1 */
  if (io_len > 1)
    bx_panic("parallel: io write from address %08x, len=%u\n",
             (unsigned) address, (unsigned) io_len);

  if ( address == BX_LPT1_IO )
      {
      BX_PAR_THIS s[0].cValue = value;
#ifdef __OS2__
      ULONG ret;
      if ( BX_PAR_THIS s[0].hPrinter )
         ret = DosWrite(BX_PAR_THIS s[0].hPrinter, &(BX_PAR_THIS s[0].cValue), 1, (ULONG *) &ret);
#else
      Bit32u ret;
      if ( BX_PAR_THIS s.fd )
         ret = ::write(BX_PAR_THIS s[0].fd, (bx_ptr_t) &(BX_PAR_THIS s[0].cValue), 1);
#endif
//      bx_printf("parallel: write %02x:%c\n", BX_PAR_THIS s[0].cValue, isprint(BX_PAR_THIS s[0].cValue) ? BX_PAR_THIS s[0].cValue : '?' );
      return;
      }

  if ( address == BX_LPT2_IO )
      {
      BX_PAR_THIS s[1].cValue = value;
#ifdef __OS2__
      ULONG ret;
      if ( BX_PAR_THIS s[1].hPrinter )
         ret = DosWrite(BX_PAR_THIS s[1].hPrinter, &(BX_PAR_THIS s[1].cValue), 1, (ULONG *) &ret);
#else
      Bit32u ret;
      if ( BX_PAR_THIS s.fd )
         ret = ::write(BX_PAR_THIS s[1].fd, (bx_ptr_t) &(BX_PAR_THIS s[1].cValue), 1);
#endif
//      bx_printf("parallel: write %02x:%c\n", BX_PAR_THIS s[1].cValue, isprint(BX_PAR_THIS s[1].cValue) ? BX_PAR_THIS s[1].cValue : '?' );
      return;
      }

  if ( address == BX_LPT3_IO )
      {
      BX_PAR_THIS s[2].cValue = value;
#ifdef __OS2__
      ULONG ret;
      if ( BX_PAR_THIS s[2].hPrinter )
         ret = DosWrite(BX_PAR_THIS s[2].hPrinter, &(BX_PAR_THIS s[2].cValue), 1, (ULONG *) &ret);
#else
      Bit32u ret;
      if ( BX_PAR_THIS s.fd )
         ret = ::write(BX_PAR_THIS s[2].fd, (bx_ptr_t) &(BX_PAR_THIS s[2].cValue), 1);
#endif
//      bx_printf("parallel: write %02x:%c\n", BX_PAR_THIS s[2].cValue, isprint(BX_PAR_THIS s[2].cValue) ? BX_PAR_THIS s[2].cValue : '?' );
      return;
      }

  if ( address == BX_LPT1_IO+2 )
      {
      if ( ( LP_PINITP | value ) == LP_PINITP )
         {
         char  cCommand;

         cCommand = 0;
#ifdef __OS2__
         ULONG ret = 1;
         ULONG lSizeCommand;
         if ( BX_PAR_THIS s[0].hPrinter )
            {
            if ( BX_PAR_THIS s[0].bPrintToFile )
               {
               ret = DosWrite(BX_PAR_THIS s[0].hPrinter, "\0x8\0xC", 2, (ULONG *) &ret);
               }
            else
               {
               ret = DosDevIOCtl(BX_PAR_THIS s[0].hPrinter, IOCTL_PRINTER, PRT_INITPRINTER,
                                 &cCommand, 1, &lSizeCommand,
                                 NULL, 0, NULL );
               }
            if ( ret == 1 )
               {
               bx_printf("PARALLEL: cannot initialize printer - spooler must be disabled!\n" );
               }
            }
#else
         Bit32u ret;
         if ( BX_PAR_THIS s[0].fd )
            ret = ::write(BX_PAR_THIS s[0].fd, (bx_ptr_t) "\0x8\0xC", 2);
#endif
//         bx_printf("parallel: init printer\n" );
         return;
         }
//      bx_printf("PARALLEL control: %x\n", value);
      }

  if ( address == BX_LPT2_IO+2 )
      {
      if ( ( LP_PINITP | value ) == LP_PINITP )
         {
         char  cCommand;

         cCommand = 0;
#ifdef __OS2__
         ULONG ret = 1;
         ULONG lSizeCommand;
         if ( BX_PAR_THIS s[1].hPrinter )
            {
            if ( BX_PAR_THIS s[1].bPrintToFile )
               {
               ret = DosWrite(BX_PAR_THIS s[1].hPrinter, "\0x8\0xC", 2, (ULONG *) &ret);
               }
            else
               {
               ret = DosDevIOCtl(BX_PAR_THIS s[1].hPrinter, IOCTL_PRINTER, PRT_INITPRINTER,
                                 &cCommand, 1, &lSizeCommand,
                                 NULL, 0, NULL );
               }
            if ( ret == 1 )
               {
               bx_printf("PARALLEL: cannot initialize printer - spooler must be disabled!\n" );
               }
            }
#else
         Bit32u ret;
         if ( BX_PAR_THIS s[1].fd )
            ret = ::write(BX_PAR_THIS s[1].fd, (bx_ptr_t) "\0x8\0xC", 2);
#endif
//         bx_printf("parallel: init printer\n" );
         return;
         }
//      bx_printf("PARALLEL control: %x\n", value);
      }

  if ( address == BX_LPT3_IO+2 )
      {
      if ( ( LP_PINITP | value ) == LP_PINITP )
         {
         char  cCommand;

         cCommand = 0;
#ifdef __OS2__
         ULONG ret = 1;
         ULONG lSizeCommand;
         if ( BX_PAR_THIS s[2].hPrinter )
            {
            if ( BX_PAR_THIS s[2].bPrintToFile )
               {
               ret = DosWrite(BX_PAR_THIS s[2].hPrinter, "\0x8\0xC", 2, (ULONG *) &ret);
               }
            else
               {
               ret = DosDevIOCtl(BX_PAR_THIS s[2].hPrinter, IOCTL_PRINTER, PRT_INITPRINTER,
                                 &cCommand, 1, &lSizeCommand,
                                 NULL, 0, NULL );
               }
            if ( ret == 1 )
               {
               bx_printf("PARALLEL: cannot initialize printer - spooler must be disabled!\n" );
               }
            }
#else
         Bit32u ret;
         if ( BX_PAR_THIS s[2].fd )
            ret = ::write(BX_PAR_THIS s[2].fd, (bx_ptr_t) "\0x8\0xC", 2);
#endif
//         bx_printf("parallel: init printer\n" );
         return;
         }
//      bx_printf("PARALLEL control: %x\n", value);
      }
}

   void
bx_parallel_c::set_status( int port, int bStatus )
{
#ifdef __OS2__
   ULONG ret;
#endif

   if ( bStatus && bx_options.lpt[port].path )
      { 
#ifdef __OS2__
      {
      ULONG ulAction = 0L;
      if ( stricmp(bx_options.lpt[port].path,"LPT1") == 0 ||
           stricmp(bx_options.lpt[port].path,"LPT2") == 0 ||
           stricmp(bx_options.lpt[port].path,"LPT3") == 0 ||
           stricmp(bx_options.lpt[port].path,"PRN") == 0 )
         {
         ret = DosOpen(bx_options.lpt[port].path, &BX_PAR_THIS s[port].hPrinter, &ulAction, 0, FILE_NORMAL, OPEN_ACTION_OPEN_IF_EXISTS, OPEN_ACCESS_READWRITE | OPEN_SHARE_DENYNONE, 0L);
         BX_PAR_THIS s[port].bPrintToFile = false;
         }
      else
         {
         ret = DosOpen(bx_options.lpt[port].path, &BX_PAR_THIS s[port].hPrinter, &ulAction, 0, FILE_NORMAL, OPEN_ACTION_CREATE_IF_NEW | OPEN_ACTION_REPLACE_IF_EXISTS, OPEN_ACCESS_READWRITE | OPEN_SHARE_DENYNONE, 0L);
         BX_PAR_THIS s[port].bPrintToFile = true;
         }
      bx_printf("PARALLEL: printer %d rc = %ld\n", port, ret );
      }
#else
      BX_PAR_THIS s[port].fd = open( bx_options.lpt[port].path, O_RDWR
#ifdef O_BINARY
                   | O_BINARY
#endif
                  );
#endif
     }
   else
     {
#ifdef __OS2__
     DosClose(BX_PAR_THIS s[port].hPrinter);
     BX_PAR_THIS s[port].hPrinter = 0;
#else
     close( BX_PAR_THIS s[port].fd );
     BX_PAR_THIS s[port].fd = 0;
#endif
     }
}
