// Copyright Notice
// ================
// BOCHS is Copyright 1994-1999 by Kevin P. Lawton.
//
// BOCHS is commercial software.
//
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
//
//     US Mail:  Kevin Lawton
//               439 Marrett Rd.
//               Lexington, MA 02421-7714
//
//     EMail:    bochs@world.std.com

#include "bochs.h"

#pragma pack(1)

static BX_OUTPUTFUNC *bxf_printf;

//#define TMSF_TRACK(time)     (*(UCHAR *)(&(time)))
//#define TMSF_MINUTE(time)    (*((UCHAR *)(&(time))+1))
//#define TMSF_SECOND(time)    (*((UCHAR *)(&(time))+2))
//#define TMSF_FRAME(time)     (*((UCHAR *)(&(time))+3))

HFILE hTestCfg;
HFILE hScreenDD;
HFILE hOemHlp;
HFILE hCDROM;

CHAR pszCdromDrive[3];
BOOL bCdromInitialized;

USHORT usDirectDisk = 0;
USHORT usTransValue = 1;
DEVICEPARAMETERBLOCK dpb = {0};
ULONG  ulHeads;

BOOL Stopped;
int BufferCount;
MCI_MIX_BUFFER *MixBuffers;			/* Device buffers          */
MCI_MIXSETUP_PARMS MixSetupParms;		/* Mixer parameters        */
MCI_BUFFER_PARMS BufferParms;			/* Device buffer parms     */
MCI_PLAY_PARMS PlayParams;
USHORT DeviceId;
ULONG ulNumBuffers;
ULONG ulBufferCount;

#define  MAX_BUFFERS    2

BOOL os2SoundSetError (APIRET rc)
{
  CHAR szErrorCode [CCHMAXPATH];

  if (rc)
    {
    mciGetErrorString (rc, (PSZ) szErrorCode, sizeof (szErrorCode));
    bxf_printf("os2 sound error rc=%ld description=%s\n", rc, szErrorCode);
    return TRUE;
    }

  return FALSE;
}

LONG APIENTRY os2SoundMixHandler (ULONG ulStatus, PMCI_MIX_BUFFER pBuffer, ULONG ulFlags)
{
  if (Stopped)
    return TRUE;

  switch (ulFlags)
    {
    case MIX_STREAM_ERROR | MIX_WRITE_COMPLETE:
      // on error, fill next buffer and continue
    case MIX_WRITE_COMPLETE:
      if (MixBuffers[ulBufferCount].ulFlags == MIX_BUFFER_EOS)
        {
        Stopped = TRUE;
        break;
        }

      ulBufferCount++;

      MixSetupParms.pmixWrite( MixSetupParms.ulMixHandle, &MixBuffers[ulBufferCount], 1 );
      break;
    }
  return TRUE;
}

void os2SoundCreate(void)
{
  Stopped = TRUE;
  DeviceId = 0;				// Clear device ID
  MixBuffers = NULL;			// Clear mixer buffers
}

void os2SoundDestroy(void)
{
  os2SoundStop ();
  os2SoundFreeBuffers ();
  os2SoundClose ();
}

BOOL os2SoundStopped(void)
{
   return Stopped;
}

void os2SoundOpen(int BitsPerSample, int SamplingRate, int DataFormat, int Channels)
{
  MCI_AMP_OPEN_PARMS AmpOpenParms = {0};
  MCI_GENERIC_PARMS GenericParms = {0};
  APIRET rc;

  if ( DeviceId )
    os2SoundDestroy ();

  DeviceId = 0;
  MixBuffers = NULL;

  AmpOpenParms.usDeviceID = (USHORT) 0;
  AmpOpenParms.pszDeviceType = (PSZ) MCI_DEVTYPE_AUDIO_AMPMIX;

  rc = mciSendCommand (0, MCI_OPEN, MCI_WAIT | MCI_OPEN_TYPE_ID | MCI_OPEN_SHAREABLE, (PVOID) &AmpOpenParms, 0);
  if (os2SoundSetError (rc))
    return;

  DeviceId = AmpOpenParms.usDeviceID;

  rc = mciSendCommand (DeviceId, MCI_ACQUIREDEVICE, MCI_EXCLUSIVE_INSTANCE, (PVOID) &GenericParms, 0);
  if (os2SoundSetError (rc))
    return;

  memset (&MixSetupParms, 0, sizeof (MCI_MIXSETUP_PARMS));
  MixSetupParms.ulBitsPerSample = BitsPerSample;
  MixSetupParms.ulSamplesPerSec = SamplingRate;
  MixSetupParms.ulFormatTag = DataFormat;
  MixSetupParms.ulChannels = Channels;
  MixSetupParms.ulFormatMode = MCI_PLAY;
  MixSetupParms.ulDeviceType = MCI_DEVTYPE_WAVEFORM_AUDIO;
  MixSetupParms.pmixEvent = os2SoundMixHandler;
  rc = mciSendCommand (DeviceId, MCI_MIXSETUP, MCI_WAIT | MCI_MIXSETUP_INIT, (PVOID) &MixSetupParms, 0);
  if (os2SoundSetError (rc))
    return;

  MixBuffers = new MCI_MIX_BUFFER [MAX_BUFFERS];

  BufferParms.ulStructLength = sizeof (BufferParms);
  BufferParms.ulNumBuffers = MAX_BUFFERS;
  BufferParms.ulBufferSize = MixSetupParms.ulBufferSize;
  BufferParms.pBufList = MixBuffers;

  rc = mciSendCommand (DeviceId, MCI_BUFFER, MCI_WAIT | MCI_ALLOCATE_MEMORY, (PVOID) &BufferParms, 0);
  if (os2SoundSetError (rc))
    return;

  BufferCount = BufferParms.ulNumBuffers;
}

int os2SoundBitsPerSample(void)
{
   return MixSetupParms.ulBitsPerSample;
}

int os2SoundSampleRate(void)
{
   return MixSetupParms.ulSamplesPerSec;
}

int os2SoundChannels(void)
{
   return MixSetupParms.ulChannels;
}

BOOL os2SoundPlay ( int length, unsigned char *data )
{
  ULONG ulIndex;
  APIRET rc;

  Stopped = FALSE;

  BufferParms.ulNumBuffers = ( length <= MixSetupParms.ulBufferSize ) ? 1 : 2;

  for( ulIndex = 0; ulIndex < BufferParms.ulNumBuffers; ulIndex++)
     {
     memset( MixBuffers[ ulIndex ].pBuffer, 0, BufferParms.ulBufferSize );
     MixBuffers[ ulIndex ].ulBufferLength = BufferParms.ulBufferSize;
     memcpy( MixBuffers[ ulIndex ].pBuffer, data, min(length,BufferParms.ulBufferSize) );
     length -= BufferParms.ulBufferSize;
     data += BufferParms.ulBufferSize;
     }

  MixBuffers[BufferParms.ulNumBuffers - 1].ulFlags = MIX_BUFFER_EOS;

  ulNumBuffers = BufferParms.ulNumBuffers;
  ulBufferCount = 0;

  rc = MixSetupParms.pmixWrite (MixSetupParms.ulMixHandle, MixBuffers, 1);

  return rc ? FALSE : TRUE;
}

void os2SoundStop (void)
{
  MCI_GENERIC_PARMS GenericParms = {0};

  if (DeviceId)
    mciSendCommand (DeviceId, MCI_STOP, MCI_WAIT, (PVOID) &GenericParms, 0);

  Stopped = TRUE;
}

void os2SoundFreeBuffers (void)
{
  if (!MixBuffers)
    return;

  for (int i = 0; i < MAX_BUFFERS; i++)
    MixBuffers [i].ulBufferLength = BufferParms.ulBufferSize;

  mciSendCommand (DeviceId, MCI_BUFFER, MCI_WAIT | MCI_DEALLOCATE_MEMORY, (PVOID) &BufferParms, 0);

  if (MixBuffers)
    delete [] MixBuffers;

  MixBuffers = NULL;
}

void os2SoundClose (void)
{
  MCI_GENERIC_PARMS GenericParms = {0};

  mciSendCommand (DeviceId, MCI_CLOSE, MCI_WAIT, (PVOID) &GenericParms, 0);
}

  void
os2DiskTranslateGeometry( unsigned short *pusCylinders, unsigned short *pusHeads )
{
   ULONG ulLocation = 0;

   if ( usTransValue == 1 )
      return;

   ulLocation = *pusHeads + *pusCylinders * ulHeads;

   *pusCylinders = ulLocation / dpb.cHeads;
   *pusHeads = ulLocation % dpb.cHeads;
}

void os2DiskCreate( char *pszPath, PUINT puiHeads, PUINT puiCylinders, PUINT puiSpt )
{
     UCHAR ucCommand;
     ULONG cbParmLenMax, cbDataLenMax;
     APIRET rc;

     rc = DosPhysicalDisk( INFO_GETIOCTLHANDLE, &usDirectDisk, sizeof(usDirectDisk), pszPath, 3 );
     if ( rc )
       {
       bxf_printf("could not open physical hard drive image file '%s' rc=%ld!\n", pszPath, rc);
       exit(0);
       }

     ucCommand = 0;
     cbParmLenMax = 0;
     memset( &dpb, 0, sizeof(DEVICEPARAMETERBLOCK) );
     cbDataLenMax = 0;

     rc = DosDevIOCtl( usDirectDisk, IOCTL_PHYSICALDISK, PDSK_GETPHYSDEVICEPARAMS, 
                       &ucCommand, 1, &cbParmLenMax,
                       &dpb, sizeof(DEVICEPARAMETERBLOCK), &cbDataLenMax);
     if ( rc )
        {
        bxf_printf( "could not get physical drive parameters rc=%ld\n", rc );
        exit(0);
        }

     bxf_printf( "physical drive parameters { cyl=%d heads=%d spt=%d }\n", dpb.cCylinders, dpb.cHeads, dpb.cSectorsPerTrack );

#ifdef CHS_TRANSLATE

     switch( dpb.cHeads )
        {
        case 255: usTransValue = 17; break;
        case 128: usTransValue = 16; break;
        case  64: usTransValue = 8;  break;
        case  32: usTransValue = 4;  break;
        case  16: usTransValue = 2;  break;
        default:  usTransValue = 1;  break;
        }

     *puiHeads     = dpb.cHeads / usTransValue;
     *puiCylinders = dpb.cCylinders * usTransValue;
     *puiSpt       = dpb.cSectorsPerTrack;

     if ( *puiHeads > 15 )
        {
        bxf_printf( "didn't translate CHS - BIOS will probably crash!\n" );
        }

#else

     usTransValue = 1;

     *puiHeads     = dpb.cHeads;
     *puiCylinders = dpb.cCylinders;
     *puiSpt       = dpb.cSectorsPerTrack;

#endif

     ulHeads = *puiHeads;

     bxf_printf( "translated drive parameters { cyl=%d heads=%d spt=%d }\n", *puiCylinders, *puiHeads, *puiSpt );
}

void os2DiskRead( long head, long cylinder, long sector, long size, char *buffer )
{
                 TRACKLAYOUT tlo;
                 ULONG cbParmLenMax, cbDataLenMax;
                 APIRET rc;

                 memset( &tlo, 0, sizeof(TRACKLAYOUT) );
                 cbParmLenMax = 0;
                 cbDataLenMax = 0;

                 tlo.bCommand = 0;

                 tlo.usHead = head;
                 tlo.usCylinder = cylinder;
                 tlo.usFirstSector = 0;
                 tlo.cSectors = 1;
                 tlo.TrackTable[0].usSectorNumber = sector;
                 tlo.TrackTable[0].usSectorSize = size;

                 os2DiskTranslateGeometry( &tlo.usCylinder, &tlo.usHead );

                 rc = DosDevIOCtl( usDirectDisk, IOCTL_PHYSICALDISK, PDSK_READPHYSTRACK, 
                       &tlo, sizeof(TRACKLAYOUT), &cbParmLenMax,
                       (bx_ptr_t) buffer, size, &cbDataLenMax);

                 if ( rc )
                    {
                    bxf_printf( "could not read physical drive rc=%ld\n", rc );
                    exit(0);
                    }
}

void os2DiskWrite( long head, long cylinder, long sector, long size, char *buffer )
{
             TRACKLAYOUT tlo;
             ULONG cbParmLenMax, cbDataLenMax;
             APIRET rc;

             memset( &tlo, 0, sizeof(TRACKLAYOUT) );
             cbParmLenMax = 0;
             cbDataLenMax = 0;

             tlo.bCommand = 0;
             tlo.usHead = head;
             tlo.usCylinder = cylinder;
             tlo.usFirstSector = 0;
             tlo.cSectors = 1;
             tlo.TrackTable[0].usSectorNumber = sector;
             tlo.TrackTable[0].usSectorSize = size;

             os2DiskTranslateGeometry( &tlo.usCylinder, &tlo.usHead );

             rc = DosDevIOCtl( usDirectDisk, IOCTL_PHYSICALDISK, PDSK_WRITEPHYSTRACK, 
                       &tlo, sizeof(TRACKLAYOUT), &cbParmLenMax,
                       (bx_ptr_t) buffer, size, &cbDataLenMax);

             if ( rc )
                {
                bxf_printf( "could not read physical drive rc=%ld\n", rc );
                exit(0);
                }
}

void cdrom_open(char *pszDrive)
{
  APIRET rc;
  ULONG ulAction;
  ULONG ulRequestLen;
  char szRequest[4] = { 'C', 'D', '0', '1' };
  ULONG ulReplyLen;
  unsigned short sectorsize;

  rc = DosOpen(pszDrive, &hCDROM, &ulAction, 0, FILE_READONLY, OPEN_ACTION_FAIL_IF_NEW|OPEN_ACTION_OPEN_IF_EXISTS, OPEN_FLAGS_FAIL_ON_ERROR|OPEN_FLAGS_DASD|OPEN_SHARE_DENYREADWRITE|OPEN_ACCESS_READONLY, 0 );
  if ( rc )
    {
    bCdromInitialized = false;
    return;
    }
  bCdromInitialized = true;
  rc = DosDevIOCtl(hCDROM,IOCTL_CDROMDISK,CDROMDISK_GETSECTORSIZE,szRequest,4,&ulRequestLen,&sectorsize,sizeof(sectorsize),&ulReplyLen);
}

void os2CdromCreate(char *pszDrive)
{
  strcpy( pszCdromDrive, pszDrive );
  cdrom_open( pszCdromDrive );
}

void os2CdromDestroy(void)
{
  if (bCdromInitialized)
    DosClose(hCDROM);
}

BOOL os2CdromInsert(void)
{
  APIRET rc;
  ULONG ulRequestLen;
  char szRequest[4] = { 'C', 'D', '0', '1' };
  ULONG ulReplyLen;
  ULONG ulReply;

  if (!bCdromInitialized)
    return false;
  rc = DosDevIOCtl(hCDROM,IOCTL_CDROMDISK,CDROMDISK_DEVICESTATUS,szRequest,4,&ulRequestLen,&ulReply,4,&ulReplyLen);
  if ( ( rc & 0xff ) == 0x10 )
    {
    os2CdromDestroy();
    cdrom_open( pszCdromDrive );
    rc = DosDevIOCtl(hCDROM,IOCTL_CDROMDISK,CDROMDISK_DEVICESTATUS,szRequest,4,&ulRequestLen,&ulReply,4,&ulReplyLen);
    }
  if (!rc && ( ulReply & ( 1 << 10 ) ) )
    return true;
  return false;
}

void os2CdromEject(void)
{
  APIRET rc;
  ULONG ulRequestLen;
  char szRequest[4] = { 'C', 'D', '0', '1' };

  if (!bCdromInitialized)
    return;
  rc = DosDevIOCtl(hCDROM,IOCTL_CDROMDISK,CDROMDISK_EJECTDISK,szRequest,4,&ulRequestLen,0,0,0);
  if ( ( rc & 0xff ) == 0x10 )
    {
    os2CdromDestroy();
    cdrom_open( pszCdromDrive );
    rc = DosDevIOCtl(hCDROM,IOCTL_CDROMDISK,CDROMDISK_EJECTDISK,szRequest,4,&ulRequestLen,0,0,0);
    }
}

BOOL os2CdromReadTOC(unsigned char* buf, int* length, BOOL msf, int start_track)
{
  APIRET rc;
  ULONG ulRequestLen;
  char szRequest[5] = { 'C', 'D', '0', '1', 0 };
  ULONG ulReplyLen;
  struct
        {
        UCHAR ucFirstTrackNumber;
        UCHAR ucLastTrackNumber;
        ULONG ulLeadOutAddress;
        } tReplyDisk;
  struct
        {
        ULONG ulAddress;
        UCHAR ucControlInformation;
        } tReplyTrack;
  int i, len;

  if (!bCdromInitialized)
    return 0;
  rc = DosDevIOCtl(hCDROM,IOCTL_CDROMAUDIO,CDROMAUDIO_GETAUDIODISK,szRequest,4,&ulRequestLen,&tReplyDisk,sizeof(tReplyDisk),&ulReplyLen);
  if ( ( rc & 0xff ) == 0x10 )
    {
    os2CdromDestroy();
    cdrom_open( pszCdromDrive );
    rc = DosDevIOCtl(hCDROM,IOCTL_CDROMAUDIO,CDROMAUDIO_GETAUDIODISK,szRequest,4,&ulRequestLen,&tReplyDisk,sizeof(tReplyDisk),&ulReplyLen);
    }
  if ( rc )
    return false;
  if ( start_track > tReplyDisk.ucFirstTrackNumber )
    return false;
  buf[2] = tReplyDisk.ucFirstTrackNumber;
  buf[3] = tReplyDisk.ucLastTrackNumber;

  len = 4;
  for ( i = start_track; i <= tReplyDisk.ucLastTrackNumber; i++ )
     {
     szRequest[4] = i;
     rc = DosDevIOCtl(hCDROM,IOCTL_CDROMAUDIO,CDROMAUDIO_GETAUDIOTRACK,szRequest,5,&ulRequestLen,&tReplyTrack,sizeof(tReplyTrack),&ulReplyLen);
     buf[len++] = 0; // Reserved
     buf[len++] = ((tReplyTrack.ucControlInformation&0x0f) << 4) | ((tReplyTrack.ucControlInformation&0xf0) >> 4) ; // ADR, control
     buf[len++] = i; // Track number
     buf[len++] = 0; // Reserved
     if (msf)
        {
        buf[len++] = 0; // reserved
        buf[len++] = TMSF_MINUTE(tReplyTrack.ulAddress);
        buf[len++] = TMSF_SECOND(tReplyTrack.ulAddress);
        buf[len++] = TMSF_FRAME(tReplyTrack.ulAddress);
        }
     else
        {
        unsigned long lba;

        // convert RedBook to HSG
        // HSG = (min * 60 + sec) * 75 + frame - 150
        lba = (TMSF_MINUTE(tReplyTrack.ulAddress) * 60 + TMSF_SECOND(tReplyTrack.ulAddress)) * 75 + TMSF_FRAME(tReplyTrack.ulAddress) - 150;

        buf[len++] = (((unsigned)lba) >> 24) & 0xff;
        buf[len++] = (((unsigned)lba) >> 16) & 0xff;
        buf[len++] = (((unsigned)lba) >> 8) & 0xff;
        buf[len++] = (((unsigned)lba) >> 0) & 0xff;
        }
     }

  *length = len;
  return true;
}

ULONG os2CdromCapacity(void)
{
  APIRET rc;
  ULONG ulRequestLen;
  char szRequest[4] = { 'C', 'D', '0', '1' };
  ULONG ulReplyLen;
  ULONG ulReply;

  if (!bCdromInitialized)
    return 0;
  rc = DosDevIOCtl(hCDROM,IOCTL_CDROMDISK,CDROMDISK_GETVOLUMESIZE,szRequest,4,&ulRequestLen,&ulReply,4,&ulReplyLen);
  if ( ( rc & 0xff ) == 0x10 )
    {
    os2CdromDestroy();
    cdrom_open( pszCdromDrive );
    rc = DosDevIOCtl(hCDROM,IOCTL_CDROMDISK,CDROMDISK_GETVOLUMESIZE,szRequest,4,&ulRequestLen,&ulReply,4,&ulReplyLen);
    }
  if ( rc )
    return 0;
  return ulReply;
}

void os2CdromReadBlock(unsigned char* buf, int lba)
{
  APIRET rc;
  ULONG ulActual;
  ULONG ulRequestLen;
  struct
        {
        UCHAR acSignature[4];
        UCHAR cAddressingMode;
        USHORT sTransferCount;
        ULONG lStartingSectorNumber;
        UCHAR cReserved;
        UCHAR cInterleaveSize;
        UCHAR cInterleaveSkipFactor;
        }
        tRequest = { { 'C', 'D', '0', '1' }, 0, 1, lba, 0, 0, 0 };
  ULONG ulReplyLen;
  struct
        {
        UCHAR acSync[12];
        UCHAR acHeader[4];
        UCHAR acDataArea[2048];
        UCHAR acEdcEcc[288];
        }
        tReply;

  if (!bCdromInitialized)
    return;

  memset( &tReply, 0, sizeof(tReply) );
  rc = DosDevIOCtl(hCDROM,IOCTL_CDROMDISK,CDROMDISK_READLONG,&tRequest,sizeof(tRequest),&ulRequestLen,&tReply,sizeof(tReply),&ulReplyLen);
  if ( ( rc & 0xff ) == 0x10 )
    {
    os2CdromDestroy();
    cdrom_open( pszCdromDrive );
    rc = DosDevIOCtl(hCDROM,IOCTL_CDROMDISK,CDROMDISK_READLONG,&tRequest,sizeof(tRequest),&ulRequestLen,&tReply,sizeof(tReply),&ulReplyLen);
    }
  if ( rc )
    bxf_printf("cdrom: read_block: READLONG(%d) returned %lx.\n", lba, rc);
  memcpy( buf, tReply.acDataArea, 2048 );
}

BOOL os2PciQueryDevice( USHORT usDeviceID, USHORT usVendorID, UCHAR ucIndex, PUCHAR pucBus, PUCHAR pucDevFn )
{
  ULONG cbParmLenMax, cbDataLenMax;
  APIRET rc;
  struct
     {
     UCHAR  ucSubFunction;
     USHORT usDeviceID;
     USHORT usVendorID;
     UCHAR  ucIndex;
     } pcifindreq;
  struct
     {
     UCHAR ucRC;
     UCHAR ucBus;
     UCHAR ucDevFunc;
     } pcifindrep;

  pcifindreq.ucSubFunction = 1;          // find pci device
  pcifindreq.usDeviceID    = usDeviceID; // pci device
  pcifindreq.usVendorID    = usVendorID; // pci vendor
  pcifindreq.ucIndex       = ucIndex;    // entry

  memset( &pcifindrep, 0, sizeof(pcifindrep) );

  cbParmLenMax = sizeof(pcifindreq);
  cbDataLenMax = sizeof(pcifindrep);

  rc = DosDevIOCtl( hOemHlp, IOCTL_OEMHLP, OEMHLP_PCI, 
                    &pcifindreq, cbParmLenMax, &cbParmLenMax,
                    &pcifindrep, cbDataLenMax, &cbDataLenMax);

  if ( pcifindrep.ucRC == 0 )
     {
     *pucBus = pcifindrep.ucBus;
     *pucDevFn = pcifindrep.ucDevFunc;

     bxf_printf("os2PciQueryDevice: device=%x vendor=%x index=%03x Bus=%x DevFunc=%x\n", (int) usDeviceID, (int) usVendorID, (int) ucIndex, (int) pcifindrep.ucBus, (int) pcifindrep.ucDevFunc );
     return TRUE;
     }

  bxf_printf("os2PciQueryDevice: device=%x vendor=%x index=%03x not found\n", (int) usDeviceID, (int) usVendorID, (int) ucIndex );
  return FALSE;
}

BOOL os2PciQueryClass( USHORT usClass, USHORT usSubClass, PUCHAR pucBus, PUCHAR pucDevFn )
{
  ULONG cbParmLenMax, cbDataLenMax;
  APIRET rc;
  struct
     {
     UCHAR  ucSubFunction;
     USHORT usSubClass;
     USHORT usClass;
     UCHAR  ucIndex;
     } pcifindreq;
  struct
     {
     UCHAR ucRC;
     UCHAR ucBus;
     UCHAR ucDevFunc;
     } pcifindrep;

  pcifindreq.ucSubFunction = 2;          // find pci device
  pcifindreq.usClass       = usClass;    // pci class
  pcifindreq.usSubClass    = usSubClass; // pci sub class
  pcifindreq.ucIndex       = 0;          // first entry

  memset( &pcifindrep, 0, sizeof(pcifindrep) );

  cbParmLenMax = cbDataLenMax = 0;

  rc = DosDevIOCtl( hOemHlp, IOCTL_OEMHLP, OEMHLP_PCI, 
                    &pcifindreq, sizeof(pcifindreq), &cbParmLenMax,
                    &pcifindrep, sizeof(pcifindrep), &cbDataLenMax);

  if ( pcifindrep.ucRC == 0 )
     {
     *pucBus = pcifindrep.ucBus;
     *pucDevFn = pcifindrep.ucDevFunc;

     bxf_printf( "os2PciQueryClass: class=%x subclass=%x bus=%x devfn=%x\n", (int) usClass, (int) usSubClass, (int) pcifindrep.ucBus, (int) pcifindrep.ucDevFunc );
     return TRUE;
     }

  bxf_printf( "os2PciQueryClass: class=%x subclass=%x not found\n", (int) usClass, (int) usSubClass );
  return FALSE;
}

void os2PciRead( UCHAR ucBus, UCHAR ucDevFn, UCHAR ucIndex, PULONG pulData, UCHAR ucSize )
{
  ULONG cbParmLenMax, cbDataLenMax;
  APIRET rc;
  struct
     {
     UCHAR ucSubFunction;
     UCHAR ucBus;
     UCHAR ucDevFunc;
     UCHAR ucCfgReg;
     UCHAR ucSize;
     } pcireadreq;
  struct
     {
     UCHAR ucRC;
     ULONG ulData;
     } pcireadrep;

  pcireadreq.ucSubFunction = 3;    // read pci device
  pcireadreq.ucBus     = ucBus;
  pcireadreq.ucDevFunc = ucDevFn;
  pcireadreq.ucCfgReg  = ucIndex;
  pcireadreq.ucSize    = ucSize;

  rc = DosDevIOCtl( hOemHlp, IOCTL_OEMHLP, OEMHLP_PCI, 
                    &pcireadreq, sizeof(pcireadreq), &cbParmLenMax,
                    &pcireadrep, sizeof(pcireadrep), &cbDataLenMax);

  *pulData = pcireadrep.ulData;

  bxf_printf("os2PciRead: bus=%x devfunc=%x index=%03x data=%08lx size=%x\n", (int) ucBus, (int) ucDevFn, (int) ucIndex, pcireadrep.ulData, (int) ucSize );
}

void os2PciWrite( UCHAR ucBus, UCHAR ucDevFn, UCHAR ucIndex, ULONG ulData, UCHAR ucSize )
{
  ULONG cbParmLenMax, cbDataLenMax;
  APIRET rc;
  struct
     {
     UCHAR ucSubFunction;
     UCHAR ucBus;
     UCHAR ucDevFunc;
     UCHAR ucCfgReg;
     UCHAR ucSize;
     ULONG ulData;
     } pciwritereq;
  struct
     {
     UCHAR ucRC;
     ULONG ulData;
     } pciwriterep;

  pciwritereq.ucSubFunction = 4;    // write pci device
  pciwritereq.ucBus     = ucBus;
  pciwritereq.ucDevFunc = ucDevFn;
  pciwritereq.ucCfgReg  = ucIndex;
  pciwritereq.ucSize    = ucSize;
  pciwritereq.ulData    = ulData;

  rc = DosDevIOCtl( hOemHlp, IOCTL_OEMHLP, OEMHLP_PCI, 
                    &pciwritereq, sizeof(pciwritereq), &cbParmLenMax,
                    &pciwriterep, sizeof(pciwriterep), &cbDataLenMax);

  bxf_printf("os2PciWrite: bus=%x devfunc=%x index=%03x data=%08lx size=%x\n", (int) ucBus, (int) ucDevFn, (int) ucIndex, ulData, (int) ucSize );
}

ULONG os2PciFindCardType(int vendor, int device)
{
   UCHAR bus, devfn;
   ULONG addr;

   if (os2PciQueryDevice( device, vendor, 0, &bus, &devfn ))
   {
      os2PciRead( bus, devfn, PCI_BASE_ADDRESS_0, &addr, sizeof(addr) );
      return addr & 0xffffff00;
   }

   return 0;
}

void os2PciReadAll( UCHAR ucBus, UCHAR ucDevFn, PCHAR pcData )
{
/*
  ULONG cbParmLenMax, cbDataLenMax;
  APIRET rc;
  struct
     {
     UCHAR ucSubFunction;
     UCHAR ucBus;
     UCHAR ucDevFunc;
     UCHAR ucCfgReg;
     UCHAR ucSize;
     } pcireadreq;
  struct
     {
     UCHAR ucRC;
     ULONG ulData;
     } pcireadrep;
  UCHAR infotbl[256];
  USHORT i;

  for ( i = 0 ; i < 256 ; i++ )
     {
     pcireadreq.ucSubFunction = 3;    // read pci device
     pcireadreq.ucBus     = ucBus;          
     pcireadreq.ucDevFunc = ucDevFn;
     pcireadreq.ucCfgReg  = i;          
     pcireadreq.ucSize    = 1;

     rc = DosDevIOCtl( hOemHlp, IOCTL_OEMHLP, OEMHLP_PCI, 
                       &pcireadreq, sizeof(pcireadreq), &cbParmLenMax,
                       &pcireadrep, sizeof(pcireadrep), &cbDataLenMax);

     infotbl[i] = pcireadrep.ulData;
     }

  memcpy( pcData, infotbl, sizeof(infotbl) );
*/
}

char *create_pmap( ULONG ulAddr, ULONG ulSize )
{
  ULONG cbParmLenMax;
  ULONG action, len;
  APIRET rc;
  struct
     {
     ULONG ulPacketLength;
     ULONG ulPhysicalAddress;
     ULONG ulMemorySize;
     ULONG ulLinearAddress;
     ULONG ulLinearFlags;
     } memmapreq;

  memmapreq.ulPacketLength = sizeof(memmapreq);
  memmapreq.ulPhysicalAddress = ulAddr;
  memmapreq.ulMemorySize = ulSize;
  memmapreq.ulLinearAddress = 0;
  memmapreq.ulLinearFlags = 0x410;

  cbParmLenMax = sizeof(memmapreq);

  rc = DosDevIOCtl( hScreenDD, SCREENDD_CATEGORY, SCREENDD_GETLINEARACCESS,
                    &memmapreq, cbParmLenMax, &cbParmLenMax,
                    0, 0, 0);

  return (char *) memmapreq.ulLinearAddress;
}

char _inb1 ( short sPort )
{
  ULONG cbParmLenMax, cbDataLenMax;
  APIRET rc;
  struct
     {
     USHORT usPort;
     USHORT usSize;
     } ioreq;
  struct
     {
     LONG lValue;
     } iorep;

  ioreq.usPort = (USHORT) sPort;
  ioreq.usSize = sizeof(CHAR);

  memset( &iorep, 0, sizeof(iorep) );

  cbParmLenMax = sizeof(ioreq);
  cbDataLenMax = sizeof(iorep);

  rc = DosDevIOCtl( hTestCfg, IOCTL_TESTCFG_SYS, TESTCFG_SYS_ISSUEINIOINSTR, 
                    &ioreq, cbParmLenMax, &cbParmLenMax,
                    &iorep, cbDataLenMax, &cbDataLenMax);

  return (CHAR) iorep.lValue;
}

short _inw1 ( short sPort )
{
  ULONG cbParmLenMax, cbDataLenMax;
  APIRET rc;
  struct
     {
     USHORT usPort;
     USHORT usSize;
     } ioreq;
  struct
     {
     LONG lValue;
     } iorep;

  ioreq.usPort = (USHORT) sPort;
  ioreq.usSize = sizeof(SHORT);

  memset( &iorep, 0, sizeof(iorep) );

  cbParmLenMax = sizeof(ioreq);
  cbDataLenMax = sizeof(iorep);

  rc = DosDevIOCtl( hTestCfg, IOCTL_TESTCFG_SYS, TESTCFG_SYS_ISSUEINIOINSTR, 
                    &ioreq, cbParmLenMax, &cbParmLenMax,
                    &iorep, cbDataLenMax, &cbDataLenMax);

  return (SHORT) iorep.lValue;
}

long _inl1 ( short sPort )
{
  ULONG cbParmLenMax, cbDataLenMax;
  APIRET rc;
  struct
     {
     USHORT usPort;
     USHORT usSize;
     } ioreq;
  struct
     {
     LONG lValue;
     } iorep;

  ioreq.usPort = (USHORT) sPort;
  ioreq.usSize = sizeof(LONG);

  memset( &iorep, 0, sizeof(iorep) );

  cbParmLenMax = sizeof(ioreq);
  cbDataLenMax = sizeof(iorep);

  rc = DosDevIOCtl( hTestCfg, IOCTL_TESTCFG_SYS, TESTCFG_SYS_ISSUEINIOINSTR, 
                    &ioreq, cbParmLenMax, &cbParmLenMax,
                    &iorep, cbDataLenMax, &cbDataLenMax);

  return iorep.lValue;
}

void _outb1 ( short sPort, char cValue )
{
  ULONG cbParmLenMax, cbDataLenMax;
  APIRET rc;
  struct
     {
     USHORT usPort;
     USHORT usSize;
     } ioreq;
  struct
     {
     LONG lValue;
     } iorep;

  ioreq.usPort = (USHORT) sPort;
  ioreq.usSize = sizeof(CHAR);

  iorep.lValue = (LONG) cValue;

  cbParmLenMax = sizeof(ioreq);
  cbDataLenMax = sizeof(iorep);

  rc = DosDevIOCtl( hTestCfg, IOCTL_TESTCFG_SYS, TESTCFG_SYS_ISSUEOUTIOINSTR, 
                    &ioreq, cbParmLenMax, &cbParmLenMax,
                    &iorep, cbDataLenMax, &cbDataLenMax);
}

void _outw1 ( short sPort, short sValue )
{
  ULONG cbParmLenMax, cbDataLenMax;
  APIRET rc;
  struct
     {
     USHORT usPort;
     USHORT usSize;
     } ioreq;
  struct
     {
     LONG lValue;
     } iorep;

  ioreq.usPort = (USHORT) sPort;
  ioreq.usSize = sizeof(SHORT);

  iorep.lValue = (LONG) sValue;

  cbParmLenMax = sizeof(ioreq);
  cbDataLenMax = sizeof(iorep);

  rc = DosDevIOCtl( hTestCfg, IOCTL_TESTCFG_SYS, TESTCFG_SYS_ISSUEOUTIOINSTR, 
                    &ioreq, cbParmLenMax, &cbParmLenMax,
                    &iorep, cbDataLenMax, &cbDataLenMax);
}

void _outl1 ( short sPort, long lValue )
{
  ULONG cbParmLenMax, cbDataLenMax;
  APIRET rc;
  struct
     {
     USHORT usPort;
     USHORT usSize;
     } ioreq;
  struct
     {
     LONG lValue;
     } iorep;

  ioreq.usPort = (USHORT) sPort;
  ioreq.usSize = sizeof(LONG);

  iorep.lValue = (LONG) lValue;

  cbParmLenMax = sizeof(ioreq);
  cbDataLenMax = sizeof(iorep);

  rc = DosDevIOCtl( hTestCfg, IOCTL_TESTCFG_SYS, TESTCFG_SYS_ISSUEOUTIOINSTR, 
                    &ioreq, cbParmLenMax, &cbParmLenMax,
                    &iorep, cbDataLenMax, &cbDataLenMax);
}

char os2VideoMode( void )
{
   char videomode;
   PPIB ppib;
   PTIB ptib;

   DosGetInfoBlocks(&ptib, &ppib);
   videomode = (ppib->pib_ultype == 0) ? BX_VIDEO_FULLSCREEN : BX_VIDEO_DIVE_WINDOW;

   if (ppib->pib_ultype == 2)
      ppib->pib_ultype = 3;

   DosSetPriority(PRTYS_THREAD, PRTYC_IDLETIME, 0, 0);

   return videomode;
}

void os2Open( BX_OUTPUTFUNC *bx_func_printf )
{
   ULONG ulAction;

   bxf_printf = bx_func_printf;

   DosOpen("OEMHLP$",  &hOemHlp,   &ulAction, 0, FILE_NORMAL, OPEN_ACTION_OPEN_IF_EXISTS, OPEN_ACCESS_READWRITE|OPEN_SHARE_DENYNONE, 0L );
   DosOpen("SCREEN$",  &hScreenDD, &ulAction, 0, FILE_NORMAL, OPEN_ACTION_OPEN_IF_EXISTS, OPEN_ACCESS_READWRITE|OPEN_SHARE_DENYNONE, 0L );
   DosOpen("TESTCFG$", &hTestCfg,  &ulAction, 0, FILE_NORMAL, OPEN_ACTION_OPEN_IF_EXISTS, OPEN_ACCESS_READWRITE|OPEN_SHARE_DENYNONE, 0L );
}

void os2Close( void )
{
   DosClose( hTestCfg );
   DosClose( hScreenDD );
   DosClose( hOemHlp );
}

